"""# lucidium.agents.base

This module provides the base class for all agents in the Lucidium framework.
"""

__all__ = ["Agent"]

from abc        import ABC, abstractmethod
from typing     import Any, List

class Agent(ABC):
    """# Abstract Agent Class
    
    This class defines the base structure for an agent that can interact with an environment. It 
    requires the implementation of methods to choose actions, save models, and load models.
    
    ## Attributes:
        * action_space  (any):  Actions possible within environment.
        * state_space   (any):  Quantification of possible states in which the environment can be 
                                observed.
        
    ## Methods:
        * select_action(state: Any) -> Any:     Choose an action based on the current state of the 
                                                environment.
        * save_model(path: str)     -> None:    Save the agent's model to the specified path.
        * load_model(path: str)     -> None:    Load the agent's model from the specified path.
    """
        
    @abstractmethod
    def act(self,
        state:  Any
    ) -> any:
        """# Act.

        Agent action is the first critical step in any reinforcement learning loop.

        Given the current state of the environment, the agent decides on an action to perform.

        ## Args:
            * state (Any):  Current environment state.

        ## Returns:
            * Any:  Chosen action.
        """
        pass
    
    @abstractmethod
    def load_model(self,
        path:   str
    ) -> None:
        """# Load Model.

        ## Args:
            * path  (str):  Path from which model save file will be loaded.
        """
        pass
    
    @abstractmethod
    def observe(self,
        new_state:  Any,
        reward:     float,
        done:       bool
    ) -> None:
        """# Observe.
        
        Agent observation is the third critical step in any reinforcement learning loop.
        
        After the agent has submitted its action, the environment will return the new state, 
        the reward, and a flag indicating if the agent has reached a terminal state. The 
        agent will use this information to update its parameters accordingly.

        ## Args:
            * new_state (Any):      State of environment after action was submitted.
            * reward    (float):    Reward yielded/penalty incurred by action submitted to 
                                    environment.
            * done      (bool):     Flag indicating if new state is terminal.
        """
        pass
    
    @abstractmethod
    def save_model(self,
        path:   str
    ) -> None:
        """# Save Model.

        ## Args:
            * path  (str):  Path to which model save file will be saved.
        """
        pass
    
    @abstractmethod
    def train_model(self,
        batch:  List[Any]
    ) -> None:
        """# Train Model.
        
        Update model using a batch of transitions or experiences.

        ## Args:
            * batch (List[Any]):    Batch of training samples.
        """
        pass


"""# lucidium.registries

This package defines the agent & environment registration systems.
"""

__all__ =   [
                # Registry & component classes.
                "Registry",
                "RegistryEntry",
                
                # Specific registries
                "AGENT_REGISTRY",
                "ENVIRONMENT_REGISTRY",
                
                # Registration decorators  
                "register_agent",
                "register_environment"
            ]

# Registry & component classes.
from .__base__      import Registry
from .entry         import RegistryEntry

# Registries.
AGENT_REGISTRY:         Registry =  Registry(name = "agents")
ENVIRONMENT_REGISTRY:   Registry =  Registry(name = "environments")

# Registration decorators 
from .decorators    import register_agent, register_environment


"""# lucidium.agents.nlm

This package defines the Neural Logic Machine and its constituent components/layers.
"""

__all__ =   [
                "NeuralLogicMachine"
            ]

from lucidium.agents.nlm.__base__   import NeuralLogicMachine


"""# lucidium.registries.decorators

Registration decorators for agents/environments.
"""

__all__ = ["register_agent", "register_environment"]

from typing import Callable, List, Optional, Type

def register_agent(
    name:   str,
    tags:   Optional[List[str]] =   [],
    parser: Optional[Callable] =    None
) -> Callable:
    """# Register Agent.

    ## Args:
        * name      (str):                  Name of agent registration entry.
        * tags      (Optional[List[str]]):  Tags that describe the agent's taxonomy.
        * parser    (Optional[Callable]):   Agent's argument registration handler.

    ## Returns:
        * Callable: Registration decorator.
    """
    # Define decorator.
    def decorator(
        cls:    Type
    ) -> Type:
        """# Registration Decorator

        ## Args:
            * cls   (Type): Agent class being registered.
        """
        # Load registry.
        from . import AGENT_REGISTRY
        
        # Register agent.
        AGENT_REGISTRY.register(
            cls =       cls,
            name =      name,
            tags =      tags,
            parser =    parser
        )
        
        # Return registered class.
        return cls
    
    # Expose decorator.
    return decorator

def register_environment(
    name:   str,
    tags:   Optional[List[str]] =   [],
    parser: Optional[Callable] =    None
) -> Callable:
    """# Register Environment.

    ## Args:
        * name      (str):                  Name of environment registration entry.
        * tags      (Optional[List[str]]):  Tags that describe the environment's taxonomy.
        * parser    (Optional[Callable]):   Environment's argument registration handler.

    ## Returns:
        * Callable: Registration decorator.
    """
    # Define decorator.
    def decorator(
        cls:    Type
    ) -> Type:
        """# Registration Decorator

        ## Args:
            * cls   (Type): Environment class being registered.
        """
        # Load registry.
        from . import ENVIRONMENT_REGISTRY
        
        # Register environment.
        ENVIRONMENT_REGISTRY.register(
            cls =       cls,
            name =      name,
            tags =      tags,
            parser =    parser
        )
        
        # Return registered class.
        return cls
    
    # Expose decorator.
    return decorator


"""# lucidium.agents.nlm.layers.permutation

Defines a dimension permutation layer for neural logic machines.
"""

__all__ = ["DimensionPermuter"]

from itertools  import permutations
from typing     import List, Tuple

from torch      import cat, Tensor
from torch.nn   import Module

class DimensionPermuter(Module):
    """# Dimension Permutation

    Creates r! new predicates by permuting the axes for r-arity predicates.

    This operation generates all permutations of the specified dimension and concatenates them along 
    the last dimension.
    
    Adapted from : https://github.com/google/neural-logic-machines/blob/master/difflogic/nn/neural_logic/modules/dimension.py
    """
    
    def __init__(self,
        dimension:      int
    ):
        """# Initialize Dimension Permutation.

        ## Args:
            * dimension (int):  Dimension along which to permute axes.
        """
        # Initialize module.
        super(DimensionPermuter, self).__init__()
        
        # Define dimension.
        self._dimension_:   int =   dimension
        
    # PROPERTIES ===================================================================================
    
    @property
    def dimension(self) -> int:
        """# Dimension (int)
        
        Dimension along which to permute axes.
        """
        return self._dimension_
    
    # METHODS ======================================================================================
        
    def forward(self,
        inputs: Tensor
    ) -> Tensor:
        """# Permute Input.
        
        Apply permutations to the input tensor along the specified dimension.

        ## Args:
            * inputs    (Tensor):   The input tensor to be permuted.

        ## Returns:
            * Tensor:   The tensor with permutations applied.
        """
        # If input dimension is one or less, permutation is not needed.
        if self.dimension <= 1: return inputs

        # Record number of dimensions in input tensor.
        dimensions:         int =           len(inputs.size())
        
        # Create index for all dimensions except for last.
        index:              Tuple[int] =    tuple(range(dimensions - 1))
        
        # Determine starting dimension for permutation.
        start_dimension:    int =           dimensions - 1 - self.dimension
        
        # Ensure starting dimension is valid.
        assert start_dimension > 0, "First dimension must be greater than zero"
        
        # Initialize list of permuted results.
        results:            List[Tensor] =  []

        # For all permutations possible in specified dimension...
        for i in permutations(index[start_dimension:]):
            
            # Make permutatiuon.
            permutation:    Tuple[int] =    index[:start_dimension] + i + (dimensions - 1,)
            
            # Apply permutation and append to results.
            results.append(inputs.permute(dims = permutation))

        # Provide concatenation of all permuted tensors along last dimension.
        return cat(tensors = results, dim = -1)
        
    def get_output_dimension(self,
        input_dimension:    int
    ) -> int:
        """# Get Output Dimension.
        
        Calculate the output dimension after permutation.

        ## Args:
            * input_dimension   (int):  The input dimension.

        ## Returns:
            * int:  The output dimension after permutation.
        """
        # Initialize multiplier.
        multiplier: int =   1
        
        # For each degree of dimension...
        for i in range(self.dimension):
            
            # Multiply by factorial of dimension.
            multiplier *=   i + 1
            
        # Provide input dimension multiplied by factorial.
        return input_dimension * multiplier


"""# lucidium.registries.base

Defines the abstract registry system.
"""

from argparse   import _SubParsersAction
from typing     import Callable, Dict, List, Optional, Type

from .entry     import RegistryEntry

class Registry():
    """# Registry
    
    Abstract registry system with lazy loading.
    """
    
    def __init__(self,
        name:   str
    ):
        """# Instantiate Registry
        
        ## Args:
            * name  (str):  Name of the sub-module that the registry represents (e.g., "agents", 
                            "environments")
        """
        # Define sub-module name.
        self._name_:    str =                       name
        
        # Initialize entry map.
        self._entries_: Dict[str, RegistryEntry] =  {}
        
        # Initialize loaded flag.
        self._loaded_:  bool =                      False
        
    # PROPERTIES ===================================================================================
    
    @property
    def entries(self) -> Optional[Dict[str, RegistryEntry]]:
        """# Registry Entries"""
        return {name: entry.cls for name, entry in self._entries_.items()}.copy()
    
    @property
    def is_loaded(self) -> bool:
        """# Registry is Loaded?"""
        return self._loaded_
    
    @property
    def name(self) -> str:
        """# Registry Name."""
        return self._name_
        
    # METHODS ======================================================================================
    
    def get_entry(self,
        key:    str
    ) -> RegistryEntry:
        """# Get Registry Entry.

        ## Args:
            * key   (str):  Key of entry being fetched.

        ## Returns:
            * RegistryEntry:    Registry entry requested.
        
        ## Raises:
            * KeyError: If registry entry does not exist.
        """
        # Ensure that registry is loaded.
        self._ensure_loaded_()
        
        # Assert that entry exists.
        if key not in self._entries_: raise KeyError(f"{key} is not registered.")
        
        # Provide requested entry.
        return self._entries_[key]
    
    def list(self,
        filter_by:  Optional[List[str]] =   []
    ) -> Optional[List[str]]:
        """# List Entries.

        ## Args:
            * filter_by (Optional[List[str]], optional):    Tags by which entries will be filtered.

        ## Returns:
            * Optional[List[str]]:  List of entry names that satisfy tag filter(s).
        """
        # Ensure that registry is loaded.
        self._ensure_loaded_()
        
        # If no filters are provided, simply return all entries.
        if not filter_by: return list(self._entries_.keys())
        
        # Otherwise, returned filtered entries.
        return  [
                    name for name, entry 
                    in self._entries_.items() 
                    if  all(
                            tag in entry.tags
                            for tag in filter_by
                        )
                ]
        
    def load_all(self) -> None:
        """# Load all Registered Modules."""
        # Simply return if registry has already been loaded.
        if self.is_loaded: return
        
        # Otherwise, import all modules...
        self._import_all_modules_()
        
        # Record that registry has now been loaded.
        self._loaded_:  bool =  True
    
    def register(self,
        cls:    Type,
        name:   str,
        tags:   Optional[List[str]] =   [],
        parser: Optional[Callable] =    None
    ) -> None:
        """# Register (Type).

        ## Args:
            * cls       (Type):                 Class being registered.
            * name      (str):                  Name of entry.
            * tags      (Optional[List[str]]):  Tags that describe the taxonomy of the class being registered.
            * parser    (Optional[Callable]):   Argument parser handler.
                                                        
        ## Raises:
            * ValueError:   If entry already exists.
        """
        # Assert that entry does not already exist.
        if name in self._entries_: raise ValueError(f"{name} if already registered.")
        
        # Register entry.
        self._entries_[name] =  RegistryEntry(
                                    cls =       cls,
                                    name =      name,
                                    tags =      tags,
                                    parser =    parser
                                )
        
    def register_parsers(self,
        parent_subparser:   _SubParsersAction
    ) -> None:
        """# Register Argument Parsers.

        ## Args:
            * parent_subparser  (_SubParsersAction):    Command sub parser of parent parser.
        """
        # Ensure that registry is loaded.
        self._ensure_loaded_()
        
        # For each registered item...
        for entry in self._entries_.values():
            
            # If entry contains argument parser...
            if entry.parser:
                
                # Register parser.
                entry.register_parser(subparser = parent_subparser)
        
    # HELPERS ======================================================================================
    
    def _ensure_loaded_(self) -> None:
        """# Ensure Registry is Loaded."""
        if not self.is_loaded: self.load_all()
        
    def _import_all_modules_(self) -> None:
        """# Import All Modules.
        
        Dynamically import all modules in the package to trigger decorators.
        """
        from importlib  import import_module
        from pkgutil    import walk_packages
        from types      import ModuleType
        
        try:# Import the main package to get its path.
            package:    ModuleType = import_module(f"lucidium.{self._name_}")
            
            # Walk through all modules in the package.
            for _, module, _ in walk_packages(
                path =      package.__path__,
                prefix =    f"lucidium.{self._name_}.",
                onerror =   lambda x: None
            ):
                try:# Attempt import of module.
                    import_module(name = module)
                    
                # Skip modules that can't be imported.
                except ImportError: continue
                   
        # Package doesn't exist or can't be imported.
        except ImportError: pass
        
    # DUNDERS ======================================================================================
    
    def __contains__(self,
        key:    str
    ) -> bool:
        """# Registry Contains Key?
        
        True if key is registered.
        """
        return key in self._entries_
    
    def __getitem__(self,
        key:    str
    ) -> RegistryEntry:
        """# Get Registry Entry.

        ## Args:
            * key   (str):  Key of entry being fetched.

        ## Returns:
            * RegistryEntry:    Registry entry requested.
        
        ## Raises:
            * KeyError: If registry entry does not exist.
        """
        # Assert that entry exists.
        if key not in self._entries_: raise KeyError(f"{key} is not registered.")
        
        # Provide requested entry.
        return self._entries_[key].cls
    
    def __len__(self) -> int:
        """# Number of Registrations."""
        return len(self._entries_)


"""# lucidium.agents.nlm.base

Implementation of Neural Logic Machine as presented in the 2019 paper by Honghua Dong et. al.

## References:
Paper:  https://arxiv.org/pdf/1904.11694.pdf
Source: https://github.com/google/neural-logic-machines
"""

from logging                        import Logger
from typing                         import Any, List, override, Optional, Tuple

from torch                          import cat, load, no_grad, save, Tensor
from torch.nn                       import Module, ModuleList

from lucidium.agents.__base__       import Agent
from lucidium.agents.nlm.__args__   import register_nlm_parser
from lucidium.agents.nlm.layers     import LogicLayer
from lucidium.registries            import register_agent
from lucidium.utilities.logger      import get_child

@register_agent(
    name =      "NLM",
    tags =      [
                    "neuro-symbolic",
                    "value-based",
                    "model-free"
                ],
    parser =    register_nlm_parser
)
class NeuralLogicMachine(Module, Agent):
    """# Neural Logic Machine

    Neural Logic Machine (NLM) is a neural-symbolic architecture for both inductive learning and 
    logic reasoning. NLMs use tensors to represent logic predicates. This is done by grounding the 
    predicate as True or False over a fixed set of objects. Based on the tensor representation, 
    rules are implemented as neural operators that can be applied over the premise tensors and 
    generate conclusion tensors.

    ## References:
    * Paper:  https://arxiv.org/pdf/1904.11694.pdf
    * Source: https://github.com/google/neural-logic-machines
    """
    
    def __init__(self,
        depth:              int,
        breadth:            int,
        input_dimensions:   Tuple[int],
        output_dimensions:  Tuple[int],
        hidden_dimension:   int,
        exclude_self:       bool =              True,
        residual:           bool =              False,
        io_residual:        bool =              False,
        recursion:          bool =              False,
        connections:        List[List[bool]] =  None
    ):
        """# Initialize Neural Logic Machine.

        ## Args:
            * depth             (int):              The number of logic layers.
            * breadth           (int):              The breadth of each logic layer.
            * input_dimensions  (Tuple[int]):       Input dimensions for each layer.
            * output_dimensions (Tuple[int]):       Output dimensions for each layer.
            * hidden_dimension  (int):              Hidden dimension of the logic model.
            * exclude_self      (bool):             Exclude multiple occurrences of the same 
                                                    variable.
            * residual          (bool):             Use residual connections.
            * io_residual       (bool):             Use input/output-only residual connections.
            * recursion         (bool):             Enable recursion for weight sharing.
            * connections       (List[List[bool]]): Connections between layers.
        """
        # Initialize module.
        super(NeuralLogicMachine, self).__init__()
        
        # Define attributes.
        self._depth_:               int =               depth
        self._breadth_:             int =               breadth
        self._residual_:            bool =              residual
        self._io_residual_:         bool =              io_residual
        self._recursion_:           bool =              recursion
        self._connections_:         List[List[bool]] =  connections
        
        # Assert that exactly one type of recursion is being requested.
        assert self._residual_ ^ self._io_residual_,    f"Only one type of residual connection type allowed."
        
        # Initialize module list.
        self._layers_:              ModuleList =        ModuleList()
        
        # Start with the input input dimension of the first layer.
        current_dimensions:         List[int] =         input_dimensions
        
        # Calculate total dimensions needed for IO residual option.
        total_output_dimension:     List[int] =         [0 for _ in range(self._breadth_ + 1)]
        
        # For each layer prescribed by depth...
        for layer in range(depth):
            
            # If IO residual option indicated, and not on first layer...
            if layer > 0 and self._io_residual_:
                
                # Update current dimensions.
                current_dimensions: list[int] =         [x + y for x, y in zip(current_dimensions, input_dimensions)]
            
            # Create a logic layer for this depth.
            layer:                  LogicLayer =        LogicLayer(
                                                            breadth =           breadth,
                                                            input_dimensions =  current_dimensions,
                                                            output_dimensions = output_dimensions,
                                                            hidden_dimension =  hidden_dimension,
                                                            exclude_self =      exclude_self,
                                                            residual =          residual
                                                        )
            
            # Update current dimensions to the output of the layer.
            current_dimensions:     Tensor =            layer._output_dimensions_
            
            # Apply masking to dimension.
            current_dimensions:     Tensor =            self._mask_(
                                                            dimensions =        current_dimensions,
                                                            index =             layer,
                                                            masked_value =      0
                                                        )
            
            # Append to layers list.
            self._layers_.append(layer)
            
        # Set output dimensions for IO residual or set to last layer's output dimensions by default.
        self._output_dimensions_:   List[int] =         total_output_dimension if io_residual else current_dimensions
        
    # PROPERTIES ===================================================================================
    
    @property
    def breadth(self) -> int:
        """# Breadth (int)
        
        The breadth of each logic layer in the Neural Logic Machine.
        """
        return self._breadth_
    
    @property
    def connections(self) -> List[List[bool]]:
        """# Connections (List[List[bool]])
        
        The connections between layers in the Neural Logic Machine.
        """
        return self._connections_
    
    @property
    def depth(self) -> int:
        """# Depth (int)
        
        The number of logic layers in the Neural Logic Machine.
        """
        return self._depth_
    
    @property
    def io_residual(self) -> bool:
        """# IO Residual (bool)
        
        Whether input/output-only residual connections are used in the Neural Logic Machine.
        """
        return self._io_residual_
    
    @property
    def layers(self) -> ModuleList:
        """# Layers (ModuleList)
        
        The list of logic layers in the Neural Logic Machine.
        """
        return self._layers_
    
    @property
    def output_dimensions(self) -> List[int]:
        """# Output Dimensions (List[int])
        
        The output dimensions of the Neural Logic Machine.
        """
        return self._output_dimensions_
    
    @property
    def recursion(self) -> bool:
        """# Recursion (bool)
        
        Whether recursion is enabled for weight sharing in the Neural Logic Machine.
        """
        return self._recursion_
    
    @property
    def residual(self) -> bool:
        """# Residual (bool)
        
        Whether residual connections are used in the Neural Logic Machine.
        """
        return self._residual_
        
    # METHODS ======================================================================================
    
    @override
    def act(self,
        state:  Any
    ) -> Any:
        """# Select Action.

        ## Args:
            * state (Any):  Observation of current environment state.

        ## Returns:
            * Any:  Action chosen.
        """
        # Ensure module is in evaluation mode.
        self.eval()
        
        # Forward pass.
        with no_grad(): return self.forward(state)
    
    def forward(self,
        inputs: List[Tensor],
        depth:  int =           None
    ) -> List[Tensor]:
        """# Forward Pass.
        
        Forward pass through Neural Logic Machine.

        ## Args:
            * inputs    (List[Tensor]):     List of input tensors.
            * depth     (int, optional):    Depth of layers to use for inference. Defaults to None.

        ## Returns:
            * list[Tensor]: Final output after logical inference.
        """
        # Initialize outputs for each group.
        outputs:    List[Tensor] =      [None for _ in range(self.breadth + 1)]
        
        # Initialize copy of input tensors.
        temp:       List[Tensor] =      inputs
        
        # If no specific depth is provided, use default.
        if depth is None:       depth:  int =   self.depth
        
        # If recursion not requested, ensure depth does not exceed the machine depth.
        if not self.recursion:  depth:  int =   min(depth, self.depth)
        
        # Define helper merge function.
        def merge(
            x:  Optional[Tensor],
            y:  Optional[Tensor]
        ) -> Tensor:
            """# Merge Tensors.

            ## Args:
                * x (Tensor | None):    Tensor input 1.
                * y (Tensor | None):    Tensor input 2.

            ## Returns:
                * Tensor:   Merged tensors.
            """
            # If x is not provided, return y.
            if x is None:   return y
            
            # If y is not provided, return x.
            if y is None:   return x
            
            # Otherwise, return concatenated tensors.
            return cat(tensors = [x, y], dim = -1)
        
        # Initialize last layer for recursion.
        last_layer: LogicLayer | None = None
        
        # For each depth level...
        for level in range(depth):
            
            # If using IO residual and not on first level...
            if level > 0 and self.io_residual:
                
                # Merge pairs of input tensors.
                for i, input in enumerate(inputs): temp[i] = merge(temp[i], input)
                
            # If recursion is enabled...
            if self._recursion_ and level >= 3:
                
                # Assert that residual is not also being used.
                assert not self.residual, "Recursion not permitted while also using residual"
                
                # Alternate between layers for weight sharing.
                last_layer, layer =     last_layer, self._layers_[level]
                
            # Otherwise, last layer is current layer.
            else: last_layer =          self._layers_[level]
            
            # Forward pass inputs through current layer.
            temp:   List[Tensor] =      last_layer(temp)
            
            # Apply mask to outputs.
            temp:   List[Tensor] =      self._mask_(
                                            dimensions = temp,
                                            index = level,
                                            masked_value = None
                                        )
            
            # If IO residual...
            if self.io_residual:
                
                # Merge outputs.
                for o, output in enumerate(temp): outputs[o] = merge(outputs[o], output)
                
        # If no IO residual, just use the final output as the result.
        if not self.io_residual: outputs = temp
        
        # Return outputs.
        return outputs
    
    @override
    def load_model(self,
        path:   str
    ) -> None:
        """# Load Agent Model.

        ## Args:
            * path  (str):  Path from which model save file can be located/loaded.
        """
        # Load model state from file.
        self.load_state_dict(load(path))
        
    @override
    def observe(self,
        new_state:  Any,
        reward:     float,
        done:       bool
    ) -> None:
        """# Observe.

        Store the transition (if needed). For NLM, this is a placeholder unless batching is used.
        """
        # NLM is stateless unless using memory buffer.
        pass
        
    @override
    def save_model(self,
        path:   str
    ) -> None:
        """# Save Agent's Model.

        ## Args:
            * path  (str):  Path at which agent's model will be saved.
        """
        # Save model state to file.
        save(self.state_dict(), path)
        
    # HELPERS ======================================================================================
        
    def _mask_(self,
        dimensions:     List[int],
        index:          int,
        masked_value:   int
    ) -> list[int]:
        """# Mask Entries.
        
        Mask out specific entries in a layer's output based on the specified connection mask.

        ## Args:
            * dimensions    (List[int]):    The dimensions of the current layer.
            * index         (int):          Index of the layer.
            * masked_value  (int):          Value to replace masked entries.

        ## Returns:
            * list[int]:    Modified dimensions list with masked entries.
        """
        # If connections have been defined...
        if self.connections is not None:
            
            # Assert that index is valid for connections.
            assert index < len(self.connections), f"Index {index} invalid for list of length {len(self.connections)}"
            
            # Reference mask for the layer's connection.
            mask:   list[bool] =    self.connections[index]
            
            # If the mask is defined...
            if mask is not None:
                
                # Assert compatible dimensions.
                assert len(mask) == len(dimensions),    f"Length of mask ({len(mask)} incompatible with dimensions length {len(dimensions)})"
                
                # Apply mask.
                dimensions: list[int] = [x if y else masked_value for x, y in zip(dimensions, mask)]
                
        # Return dimensions.
        return dimensions


"""# lucidium.utilities

This package defines various utilities used throughout the application.
"""

__all__ =   [
                "BANNER",
                "get_child",
                "get_logger"
            ]

from lucidium.utilities.banner  import BANNER
from lucidium.utilities.logger  import get_child, get_logger


"""# lucidium.main

Lucidium application driver.
"""

from argparse           import Namespace
from logging            import Logger

from lucidium.__args__  import parse_lucidium_arguments
from lucidium.utilities import *

def main(*args, **kwargs) -> None:
    """# Execute Application."""
    # Parse arguments.
    _arguments_:    Namespace = parse_lucidium_arguments()
    
    # Initialize logger.
    _logger_:       Logger =    get_logger(
                                    logger_name =   "lucidium",
                                    logging_level = _arguments_.logging_level,
                                    logging_path =  _arguments_.logging_path
                                )
    
    try:# Log banner.
        _logger_.info(BANNER)
    
    # Catch wildcard errors.
    except Exception as e:  _logger_.critical(f"Unexpected error: {e}", exc_info = True)
    
    # Exit gracefully.
    finally:                _logger_.debug("Exiting")

if __name__ == "__main__": main()


"""# lucidium.agents.nlm.layers.logic

Defines a logic layer that performs differentiable logic deduction.
"""

__all__ = ["LogicLayer"]

from logging                                import Logger
from typing                                 import Any, List, Union

from torch                                  import cat, Tensor
from torch.nn                               import Module, ModuleList

from lucidium.agents.nlm.layers.expansion   import DimensionExpander
from lucidium.agents.nlm.layers.inference   import InferenceLayer
from lucidium.agents.nlm.layers.permutation import DimensionPermuter
from lucidium.agents.nlm.layers.reduction   import DimensionReducer
from lucidium.utilities                     import get_child

class LogicLayer(Module):
    """# Logic Layer
    
    A logic layer that performs differentiable logic deduction.
    
    Each logic layer computes inference over input predicates by applying logic rules. The logic is 
    performed in two stages: 
    * Expansion/Reduction:      For inter-group deduction
    * Logic model application:  For intra-group deduction
    
    Adapted from: https://github.com/google/neural-logic-machines/blob/master/difflogic/nn/neural_logic/layer.py
    """
    
    def __init__(self,
        breadth:            int,
        input_dimensions:   List[int],
        output_dimensions:  List[int],
        hidden_dimension:   int,
        exclude_self:       bool =      True,
        residual:           bool =      False
    ):
        """# Initialize Logic Layer.

        ## Args:
            * breadth           (int):          The breadth (max order) of the logic layer.
            * input_dimensions  (List[int]):    The number of input channels for each input group.
            * output_dimensions (List[int]):    The number of output channels for each group.
            * hidden_dimension  (int):          The hidden dimension of the logic model.
            * exclude_self      (bool):         Whether to exclude multiple occurrences of the same 
                                                variable in logic deduction.
            * residual          (bool):         Whether to use residual connections for the outputs.
        """
        # Initialize module.
        super(LogicLayer, self).__init__()
        
        # Initialize logger.
        self.__logger__:                Logger =            get_child("logic-layer")
        
        # Assert that breadth is positive.
        assert 0 < breadth, f"Breadth expected to be positive, got {breadth}"
        
        # Warn of performance complications for breadth greater than 3.
        if breadth > 3: self.__logger__.warning(f"Using a breadth greater than 3 may cause performance complications (speed & memory).")
        
        # Define maximum order (degree) of logic.
        self._max_order_:               int =               breadth
        
        # Set residual flag.
        self._residual_:                bool =              residual
        
        # Ensure input and output dimensions are lists of the correct length, being no greater than 
        # max order.
        input_dimensions:               List[int] =             self._make_list_(
                                                                    x =     input_dimensions, 
                                                                    n =     self._max_order_ + 1,
                                                                    dtype = int
                                                                )
        output_dimensions:              List[int] =             self._make_list_(
                                                                    x =     output_dimensions,
                                                                    n =     self._max_order_ + 1,
                                                                    dtype = int
                                                                )
        
        # Initialize module lists.
        self._logic_:                   ModuleList =            ModuleList()
        self._dimension_permuters_:     ModuleList =            ModuleList()
        self._dimension_expanders_:     ModuleList =            ModuleList()
        self._dimension_reducers_:      ModuleList =            ModuleList()
        
        # For each order of logic prescribed...
        for order in range(self._max_order_ + 1):
            
            # Start with input dimension.
            current_dimension:          int =                   input_dimensions[order]
            
            # For all but the first order...
            if order > 0:
                
                # Initialize expander.
                expander:               DimensionExpander =     DimensionExpander(
                                                                    dimension = order - 1
                                                                )
                
                # Append to expanders list.
                self._dimension_expanders_.append(expander)
                
                # Add expanded dimension.
                current_dimension +=                            expander.get_output_dimension(
                                                                    input_dimension = input_dimensions[order - 1]
                                                                )
            
            # Otherwise, no expansion is needed for first degree.
            else: self._dimension_expanders_.append(None)
            
            # For the next degree...
            if order + 1 < self._max_order_ + 1:
                
                # Define a dimension reducer.
                reducer:                DimensionReducer =      DimensionReducer(
                                                                    dimension =     order + 1,
                                                                    exclude_self =  exclude_self
                                                                )
                
                # Append to reducers list.
                self._dimension_reducers_.append(reducer)
                
                # Add reduced dimension.
                current_dimension +=                            reducer.get_output_dimension(
                                                                    input_dimension = input_dimensions[order + 1]
                                                                )
            
            # Otherwise, no reduction is needed for first degree.
            else: self._dimension_reducers_.append(None)
            
            # For the first order...
            if current_dimension == 0:
                
                # Do no create logic or permutation.
                self._dimension_permuters_.append(None)
                self._logic_.append(None)
                
                # Set output dimension to zero when no logic is applied.
                output_dimensions[order] =                      0
                
            # Otherwise...
            else:
                
                # Create permutation layer for this group.
                permutation:            DimensionPermuter = DimensionPermuter(
                                                                dimension = order
                                                            )
                
                # Append to permutation list.
                self._dimension_permuters_.append(permutation)
                
                # Apply permutation to current dimensions.
                current_dimension:      int =               permutation.get_output_dimension(
                                                                input_dimension =   current_dimension
                                                            )
                
                # Append logic layer.
                self._logic_.append(InferenceLayer(
                    input_dimension =   current_dimension,
                    output_dimension =  output_dimensions[order],
                    hidden_dimension =  hidden_dimension,
                    activation =        True
                ))
                
        # Define dimensions lists.
        self._input_dimensions_:        List[int] =             input_dimensions
        self._output_dimensions_:       List[int] =             output_dimensions
        
        # If residual connections is specified...
        if self._residual_:
            
            # For each input dimension, add input dimensions to output dimensions for residual.
            for i in range(len(self._input_dimensions_)): self._output_dimensions_[i] += self._input_dimensions_[i]
            
    # PROPERTIES ===================================================================================
    
    @property
    def breadth(self) -> int:
        """# Breadth (int)
        
        The maximum order of the logic layer.
        """
        return self._max_order_
    
    @property
    def exclude_self(self) -> bool:
        """# Exclude Self (bool)
        
        Whether to exclude multiple occurrences of the same variable in logic deduction.
        """
        return self._residual_
    
    @property
    def expanders(self) -> ModuleList:
        """# Dimension Expanders (ModuleList)
        
        List of dimension expanders for each group of the logic layer.
        """
        return self._dimension_expanders_
    
    @property
    def input_dimensions(self) -> List[int]:
        """# Input Dimensions (List[int])
        
        Number of input channels for each group of the logic layer.
        """
        return self._input_dimensions_
    
    @property
    def logic(self) -> ModuleList:
        """# Logic Inference Layers (ModuleList)
        
        List of logic inference layers for each group of the logic layer.
        """
        return self._logic_
    
    @property
    def max_order(self) -> int:
        """# Max Order (int)
        
        The maximum order of the logic layer.
        """
        return self._max_order_
    
    @property
    def output_dimensions(self) -> List[int]:
        """# Output Dimensions (List[int])
        
        Number of output channels for each group of the logic layer.
        """
        return self._output_dimensions_
    
    @property
    def permuters(self) -> ModuleList:
        """# Dimension Permutations (ModuleList)
        
        List of dimension permutations for each group of the logic layer.
        """
        return self._dimension_permuters_
    
    @property
    def reducers(self) -> ModuleList:
        """# Dimension Reducers (ModuleList)
        
        List of dimension reducers for each group of the logic layer.
        """
        return self._dimension_reducers_
    
    @property
    def residual(self) -> bool:
        """# Residual (bool)
        
        Whether to use residual connections for the outputs.
        """
        return self._residual_
            
    # METHODS ======================================================================================
            
    def forward(self,
        inputs: List[Tensor]
    ) -> List[Tensor]:
        """# Forward Pass.
        
        Forward pass through logic layer.

        ## Args:
            * inputs    (list[Tensor]): List on inputs for each group of the logic layer.

        ## Returns:
            * list[Tensor]: Output from each logic layer group.
        """
        # Assert that the number of inputs matches expected order.
        assert len(inputs) == self.max_order + 1, f"Number of input tensors ({len(inputs)} should be {self.max_order})"
        
        # Initialize list to store outputs.
        outputs:        List[Tensor] =  []
        
        # For each group of logic layer...
        for group in range(self.max_order + 1):
            
            # Initialize temporary list to store input of adjacent groups.
            temp:       List[Tensor] =  []
            
            # If not the first group...
            if group > 0 and self.input_dimensions[group - 1] > 0:
                
                # Reference expansion size.
                n:      int =           inputs[group].size(1) if group == 1 else None
                
                # Append expanded tensor to temporary list.
                temp.append(self.expanders[group](inputs[group -1], n))
                
            # If not the last group, and to temporary list.
            if group < len(inputs) and self.input_dimensions[group] > 0: temp.append(inputs[group])
            
            # Reduce inputs from next group.
            if group + 1 < len(inputs) and self.input_dimensions[group + 1] > 0: temp.append(self.reducers[group](inputs[group + 1]))
            
            # If no inputs were collected, output should be None.
            if len(temp) == 0: output: Tensor = None
            
            # Otherwise...
            else:
                
                # Concatenate all inputs along the last dimension.
                temp =                  cat(tensors = temp, dim = -1)
                
                # Apply permutation to the concatenated inputs.
                temp =                  self.permuters[group](temp)
                
                # Apply logic inference layer.
                output: Tensor =        self.logic[group](temp)
                
            # If residual is specified...
            if self.residual and self.input_dimensions[group] > 0:
                
                # Concatenate the input with the output.
                output: Tensor =        cat(tensors = [inputs[group], output], dim = -1)
                
            # Append output to list.
            outputs.append(output)
            
        # Return final output list.
        return outputs
    
    # HELPERS ======================================================================================
        
    def _make_list_(self,
        x:      Union[List, Any],
        n:      int,
        dtype:  type
    ) -> List[any]:
        """# Make List of N Elements.
        
        Ensure that `x` is either a list of `n` elements or single element of type `dtype`.

        ## Args:
            * x     (List | Any):   Element or list.
            * n     (int):          Desired length of list.
            * dtype (type):         Expected type of list elements.

        ## Returns:
            * list[any]:    List of length `n` and type `dtype`.
        """
        # Assert that data type is not list.
        assert dtype is not list,               f"Data type should not be list."
        
        # If `x` is a single element of type `dtype`, create an `n` length list of it.
        if isinstance(x, dtype): x = [x,] * n
        
        # Assert that list is desired length.
        assert len(x) == n,                     f"Parameters should be {dtype} or list of {n} elements."
        
        # For each element in list...
        for e, element in enumerate(x):
            
            # Assert that it is of proper type.
            assert isinstance(element, dtype),  f"Elements expected to be {dtype}, but x[{e}] is {type(element)}"
            
        # Return list.
        return x


"""# lucidium.agents.nlm.args

Argument definitions and parsing for Neural Logic Machine agent.
"""

__all__ = ["register_nlm_parser"]

from argparse   import _ArgumentGroup, ArgumentParser, _SubParsersAction

def register_nlm_parser(
    parent_subparser:   _SubParsersAction
) -> None:
    """# Register Neural Logic Machine Argument Parser.
    
    Add NLM agent sub-parser & arguments to parent's sub-parser.

    ## Args:
        * parent_subparser  (_SubParsersAction):    Parent's sub-parser object.
    """
    # Initialize Neural Logic Machine agent parser.
    _parser_:           ArgumentParser =    parent_subparser.add_parser(
        name =          "nlm",
        help =          "Neural Logic Machine",
        description =   """Neural Logic Machine (NLM) is a neural-symbolic architecture for both 
                        inductive learning and logic reasoning. NLMs use tensors to represent logic 
                        predicates. This is done by grounding the predicate as True or False over a 
                        fixed set of objects. Based on the tensor representation, rules are 
                        implemented as neural operators that can be applied over the premise 
                        tensors and generate conclusion tensors.""",
        epilog =        """Implementation based on "Neural Logic Machines" by Honghua Dong et. al. 
                        (2019). Link to paper: https://arxiv.org/pdf/1904.11694.pdf"""
    )
    
    # Define sub-parser.
    _subparser_:        _SubParsersAction = _parser_.add_subparsers(
        title =         "action",
        description =   """Neural Logic Machine commands.""",
        dest =          "action"
    )

    # +============================================================================================+
    # | BEGIN ARGUMENTS                                                                            |
    # +============================================================================================+
    
    # DIMENSIONALITY ===============================================================================
    _dimensionality_:   _ArgumentGroup =    _parser_.add_argument_group(title = "Dimensionality")
    
    _dimensionality_.add_argument(
        "--depth",
        dest =          "depth",
        type =          int,
        default =       3,
        help =          """Number of logic layers with which NLM will be initialized. Defaults to 
                        3."""
    )
    
    _dimensionality_.add_argument(
        "--breadth",
        dest =          "breadth",
        type =          int,
        default =       2,
        help =          """Breadth with which each logic layer will be initialized. Defaults to 
                        2."""
    )
    
    _dimensionality_.add_argument(
        "--hidden-dimension",
        dest =          "hidden_dimension",
        type =          int,
        default =       128,
        help =          """Dimension of hidden layers within logic layers. Defaults to 128."""
    )
    
    # FEATURE FLAGS ================================================================================
    _flags_:            _ArgumentGroup =    _parser_.add_argument_group(title = "Feature Flags")
    
    _flags_.add_argument(
        "--exclude-self",
        dest =          "exclude_self",
        action =        "store_true",
        default =       False,
        help =          """Exclude multiple occurrences of the same variable."""
    )
    
    _flags_.add_argument(
        "--io-residual",
        dest =          "io_residual",
        action =        "store_true",
        default =       False,
        help =          """Use input/output-only residual connections. Mututally exclusive to using 
                        "--residual" flag."""
    )
    
    _flags_.add_argument(
        "--residual",
        dest =          "residual",
        action =        "store_true",
        default =       False,
        help =          """Use fully residual connections. Mututally exclusive to using 
                        "--io-residual" flag."""
    )
    
    _flags_.add_argument(
        "--recursion",
        dest =          "recursion",
        action =        "store_true",
        default =       False,
        help =          """Enable recursion for weight sharing."""
    )

    # +============================================================================================+
    # | END ARGUMENTS                                                                              |
    # +============================================================================================+


"""# lucidium.agents.nlm.layers.mlp

Defines a multi-layer perceptron (MLP) layer with activation for neural logic machines.
"""

__all__ = ["MLPLayer"]

from typing                             import List, Union

from torch                              import Tensor
from torch.nn                           import Linear, Module, Sequential

from lucidium.agents.nlm.layers.linear  import LinearLayer

class MLPLayer(Module):
    """# MLP Layer
    
    A multi-layer perceptron (MLP) layer with activation for hidden layers.

    Adapted from jatorch.nn.cnn.layers: https://github.com/vacancy/Jacinle/blob/master/jactorch/nn/cnn/layers.py
    """
    
    def __init__(self,
        input_dimension:    int,
        output_dimension:   int,
        hidden_dimensions:  Union[List[int], int]
    ):
        """# Initialize MLP Layer.

        ## Args:
            * input_dimension   (int):              Number of input features.
            * output_dimension  (int):              Number of output features.
            * hidden_dimensions (List[int] | int):  List of hidden layer sizes.
        """
        # Initialize module.
        super(MLPLayer, self).__init__()
        
        # Define dimensions.
        self._input_dimension_:     int =           input_dimension
        self._output_dimension_:    int =           output_dimension
        self._hidden_dimensions_:   List[int] =     hidden_dimensions if isinstance(hidden_dimensions, list) else [hidden_dimensions]
        
        # Define attributes.
        self._flatten_:             bool =          True
        self._last_activation_:     bool =          False
        
        # Define dimensions list.
        dimensions:                 List[int] =     [self._input_dimension_] + self._hidden_dimensions_ + [self._output_dimension_]
        
        # Initialize modules list.
        modules:                    List[Module] =  []
        
        # or each hidden dimension...
        for i in range(len(self._hidden_dimensions_)):
            
            # Create hidden layer.
            modules.append(LinearLayer(
                in_features =   dimensions[i],
                out_features =  dimensions[i + 1]
            ))
            
        # Append final layer.
        modules.append(Linear(
            in_features =   dimensions[-2],
            out_features =  dimensions[-1],
            bias =          True
        ))
        
        # Apply layers sequentially.
        self._mlp_:                 Sequential =    Sequential(*modules)
        
    # PROPERTIES ===================================================================================
    
    @property
    def flatten(self) -> bool:
        """# Flatten (bool)
        
        Whether to flatten the input before passing it through the MLP.
        """
        return self._flatten_
    
    @property
    def hidden_dimensions(self) -> List[int]:
        """# Hidden Dimensions (List[int])
        
        List of hidden layer sizes.
        """
        return self._hidden_dimensions_
    
    @property
    def input_dimension(self) -> int:
        """# Input Dimension (int)
        
        Dimension of input features.
        """
        return self._input_dimension_
    
    @property
    def last_activation(self) -> bool:
        """# Last Activation (bool)
        
        Whether to apply an activation function after the last layer.
        """
        return self._last_activation_
    
    @property
    def mlp(self) -> Sequential:
        """# MLP (Sequential)
        
        The sequential module containing the MLP layers.
        """
        return self._mlp_
    
    @property
    def output_dimension(self) -> int:
        """# Output Dimension (int)
        
        Dimension of output features.
        """
        return self._output_dimension_
    
    # METHODS ======================================================================================

    def forward(self,
        input:  Tensor
    ) -> Tensor:
        """# Forward Pass.
        
        Forward pass through the MLP.

        ## Args:
            * input (Tensor):   The input tensor.

        ## Returns:
            * Tensor:   The output tensor after passing through the MLP layers.
        """
        # Flatten if desired.
        if self.flatten: input: Tensor = input.view(input.size(0), -1)
        
        # Make forward pass through layer.
        return self.mlp(input)
    
    def reset_parameters(self) -> None:
        """# Reset Parameters.
        
        Reset parameters of linear layer.
        """
        # For each module in layer...
        for module in self.modules():
            
            # If it is a linear layer, reset the parameters.
            if isinstance(module, Linear): module.reset_parameters()


"""# lucidium.agents.nlm.layers

This package defines the various layer components of the Neural Logic Machine.
"""

__all__ =   [
                "DimensionExpander",
                "DimensionPermuter",
                "DimensionReducer",
                "LinearLayer",
                "LogicLayer",
                "MLPLayer"
            ]

from lucidium.agents.nlm.layers.expansion   import DimensionExpander
from lucidium.agents.nlm.layers.permutation import DimensionPermuter
from lucidium.agents.nlm.layers.reduction   import DimensionReducer
from lucidium.agents.nlm.layers.linear      import LinearLayer
from lucidium.agents.nlm.layers.logic       import LogicLayer
from lucidium.agents.nlm.layers.mlp         import MLPLayer


"""# lucidium.environments.tic_tac_toe.args

Argument definitions and parsing for Tic-Tac-Toe environment.
"""

__all__ = ["register_tic_tac_toe_parser"]

from argparse   import _ArgumentGroup, ArgumentParser, _SubParsersAction

def register_tic_tac_toe_parser(
    parent_subparser:   _SubParsersAction
) -> None:
    """# Register Tic-Tac-Toe Argument Parser.

    ## Args:
        * parent_subparser  (_SubParsersAction):    Parent's sub-parser object.
    """
    # Initialize parser.
    _parser_:           ArgumentParser =    parent_subparser.add_parser(
        name =          "tic-tac-toe",
        help =          "Tic-Tac-Toe environment",
        description =   """Tic-Tac-Toe is a two-player game where players take turns marking a 
                        square in a 3x3 grid. The first player to align three of their marks 
                        horizontally, vertically, or diagonally wins the game. This environment 
                        provides a simple but strategic domain for testing neuro-symbolic 
                        reinforcement learning approaches.""",
        epilog =        """Example usage of creating a 4x4 Tic-Tac-Toe game with custom reward 
                        structure: lucidium tictactoe --size 4 --win-reward 2.0 
                        --invalid-move-penalty -1.0"""
    )
    
    # Define sub-parser for actions (train, play, analyze, etc.)
    _subparser_:        _SubParsersAction = _parser_.add_subparsers(
        title =         "action",
        description =   "Tic-Tac-Toe environment commands.",
        dest =          "action"
    )

    # +============================================================================================+
    # | BEGIN ARGUMENTS                                                                            |
    # +============================================================================================+
    
    # BOARD CONFIGURATION ==========================================================================
    _configuration_:    _ArgumentGroup =    _parser_.add_argument_group(
        title =         "Board Configuration",
        description =   """Basic game setup parameters."""
    )
    
    _configuration_.add_argument(
        "--size",
        dest =          "size",
        type =          int,
        default =       3,
        choices =       range(3, 11),
        help =          """Size of (square) game board. Must be between 3 and 10. Defaults to 3 for 
                        standard tic-tac-toe."""
    )
    
    # REWARD/PENALTY ===============================================================================
    _rewards_:          _ArgumentGroup =    _parser_.add_argument_group(
        title =         "Rewards/Penalties",
        description =   """Configure reward/penalty values."""
    )
    
    _rewards_.add_argument(
        "--win-reward",
        dest =          "win_reward",
        type =          float,
        default =       1.0,
        help =          """Reward yielded by winning the game. Defaults to 1.0."""
    )
    
    _rewards_.add_argument(
        "--loss-penalty",
        dest =          "loss_penalty",
        type =          float,
        default =       -1.0,
        help =          """Penalty incurred by losing the game. Defaults to -1.0."""
    )
    
    _rewards_.add_argument(
        "--draw-penalty",
        dest =          "draw_penalty",
        type =          float,
        default =       -0.0,
        help =          """Reward yielded/penalty incurred when game ends in a draw (Neither player 
                        wins). Defaults to -0.0."""
    )
    
    _rewards_.add_argument(
        "--invalid-move-penalty",
        dest =          "invalid_move_penalty",
        type =          float,
        default =       -0.5,
        help =          """Penalty incurred for attempting an invalid move. Defaults to -0.5."""
    )
    
    _rewards_.add_argument(
        "--move-penalty",
        dest =          "move_penalty",
        type =          float,
        default =       -0.1,
        help =          """Cost of making any move during the game. Defaults to -0.1."""
    )

    # +============================================================================================+
    # | END ARGUMENTS                                                                              |
    # +============================================================================================+


"""# lucidium.environments.tic_tac_toe.components.board

Implementation of board component for tic-tac-toe.
"""

from typing                                             import List, Tuple, Union

from torch                                              import Tensor

from lucidium.environments.tic_tac_toe.components.cell  import Cell

class Board():
    """# (Tic-Tac-Toe) Board
    
    Board representation and utilities.
    """
    
    def __init__(self,
        size:   int =   3
    ):
        """# Instanstiate Board.

        ## Args:
            * size  (int):  Dimension of (square) board. Maximum is 10. Defaults to 3.
        """
        # Assert that size is in range 3 - 10.
        assert 3 <= size <= 10, f"Board size must be 3 - 10, got {size}"
        
        # Define properties.
        self._size_:    int =   size
        
        # Instantiate cells.
        self._grid_:    List[List[Cell]] =  [
                                                [
                                                    Cell(coordinate = (r, c)) 
                                                    for c in range(self._size_)
                                                ]
                                                for r in range(self._size_)
                                            ]
        
    # PROPERTIES ===================================================================================
    
    @property
    def available_moves(self) -> List[Tuple[int, int]]:
        """# Available Moves

        Coordinates of currently available moves.
        """
        return [(r, c) for r in range(self._size_) for c in range(self._size_) if self._grid_[r][c].is_empty]
    
    @property
    def has_winner(self) -> bool:
        """# (Game) has Winner?
        
        True if a player has won the game.
        """
        return any(self._line_is_a_win_(line) for line in self.lines)
    
    @property
    def is_full(self) -> bool:
        """# (Board) is Full?

        ## Returns:
            * bool: True if no empty cells remain.
        """
        return all(not cell.is_empty for row in self._grid_ for cell in row)
    
    @property
    def lines(self) -> List[List[Cell]]:
        """# (Board) Lines

        Cell groupings for each row, column, & diagonal.
        """
        return  [[self._grid_[r][c] for c in range(self._size_)] for r in range(self._size_)]   +\
                [[self._grid_[c][r] for c in range(self._size_)] for r in range(self._size_)]   +\
                [[self._grid_[i][i] for i in range(self._size_)]]                               +\
                [[self._grid_[i][self._size_ - i - 1] for i in range(self._size_)]]
        
    # METHODS ======================================================================================
    
    def mark(self,
        row:    int,
        column: int,
        entry:  Union[str, int, Tensor]
    ) -> bool:
        """# Mark Cell.

        ## Args:
            * row       (int):                      Row containing cell to mark.
            * column    (int):                      Column containing cell to mark.
            * entry     (Union[str, int, Tensor]):  Player entry representation:
                * Symbol ("X", "O", " ")
                * Number (1, -1, 0)
                * One-hot tensor (e.g., [1, 0, 0])
                
        ## Returns:
            * bool: True if cell was marked.
        """
        return self._grid_[row][column].mark(entry = entry)
    
    def mark_by_index(self,
        index:  int,
        entry:  Union[str, int, Tensor]
    ) -> bool:
        """# Mark Cell by Index.

        ## Args:
            * index (int):                      Index of cell to mark
            * entry (Union[str, int, Tensor]):  Player entry representation:
                * Symbol ("X", "O", " ")
                * Number (1, -1, 0)
                * One-hot tensor (e.g., [1, 0, 0])
                
        ## Returns:
            * bool: True if cell was marked.
        """
        return self.mark(*self._index_to_coordinate_(index = index), entry = entry)
    
    def reset(self) -> None:
        """# Reset (Board).
        
        Reset board and all contained cells to their initial states.
        """
        # For each row...
        for row in self._grid_:
            
            # Reset each cell in row.
            for cell in row: cell.reset()
            
    # HELPERS ======================================================================================
    
    def _coordinate_to_index_(self,
        coordinate: Tuple[int, int]
    ) -> int:
        """# (Convert) to Index.
        
        Convert cell coordinate to board index.

        ## Args:
            * coordinate    (Tuple[int, int]):  Row, column coordinate being converted.

        ## Returns:
            * int:  Index equivalent of coordinate provided.
        """
        return coordinate[0] * self._size_ + coordinate[1]
    
    def _index_to_coordinate_(self,
        index:  int
    ) -> Tuple[int, int]:
        """# (Convert) to Coordinate.
        
        Convert board index to cell coordinate.

        ## Args:
            * index (int):  Board index being converted.

        ## Returns:
            * Tuple[int, int]:  Coordinate equivalent of index provided.
        """
        return divmod(index, self._size_)
    
    def _line_is_a_win_(self,
        line:   List[Cell]
    ) -> bool:
        """# Line is a Win?

        ## Args:
            * line  (List[Cell]):   Cell grouping being verified.

        ## Returns:
            * bool: True if line if occupied by one player.
        """
        # If any cells are empty, there is no win.
        if any(cell.is_empty for cell in line): return False
        
        # Otherwise, indicate that only one player occupies all cells.
        return len({cell.entry for cell in line}) == 1
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation

        Object representation of board.
        """
        return f"<Board(size = {self._size_})>"
    
    def __str__(self) -> str:
        """# String Representation

        String representation of board.
        """
        # Predefine separation line and column index.
        line:           str =   ("\n   " + ("───┼" * (self._size_ - 1)) + ("─" * 3))
        grid_string:    str =   ("   ") + " ".join(f" {column} " for column in range(self._size_))
        
        # For each row in grid...
        for r, row in enumerate(self._grid_):
            
            # Render cell row.
            grid_string += f"\n {r} " + "│".join(f" {cell} " for cell in row)
            
            # Append line if not on last row.
            grid_string += line if r < self._size_ - 1 else ""
            
        # Return string representation.
        return grid_string


"""# lucidium.agents.nlm.layers.expansion

Defines the dimensional expansion layer for neural logic machines.
"""

__all__ = ["DimensionExpander"]

from torch                          import Tensor
from torch.nn                       import Module

from lucidium.agents.nlm.__util__   import broadcast

class DimensionExpander(Module):
    """# Dimension Expander

    Captures a free variable into predicates, implemented by broadcast.
    
    Adapted from : https://github.com/google/neural-logic-machines/blob/master/difflogic/nn/neural_logic/modules/dimension.py
    """
    
    def __init__(self,
        dimension:  int
    ):
        """# Initialize Dimension Expander.

        ## Args:
            * dimension (int):  Dimension along which expansion will occur.
        """
        # Initialize module.
        super(DimensionExpander, self).__init__()
        
        # Define dimension.
        self._dimension_:   int =   dimension
        
    # PROPERTIES ===================================================================================
    
    @property
    def dimension(self) -> int:
        """# Dimension (int)
        
        Dimension along which expansion will occur.
        """
        return self._dimension_
    
    # METHODS ======================================================================================
        
    def forward(self,
        inputs: Tensor,
        n:      int | None =    None
    ) -> Tensor:
        """# Expand Input.
        
        Expand input tensor by unsqueezing it along the specified dimension.

        ## Args:
            * inputs    (Tensor):               Input tensor being expanded.
            * n         (int | None, optional): Size to which dimension will be expanded. Required 
                                                if dimension is zero. Defaults to None.

        ## Returns:
            * Tensor:   Expanded tensor.
        """
        # If dimension is zero...
        if self.dimension == 0:
            
            # Assert that `n` is provided.
            assert n is not None,   f"N parameter must be provided when dimension is zero."
            
        # Otherwise, default to size of respective input dimension.
        n:          int =   inputs.size(dim = self.dimension) if n is None else n
        
        # Add extra dimension for broadcasting.
        dimension:  int =   self.dimension + 1
        
        # Return expanded tensor.
        return  broadcast(
                    tensor =    inputs.unsqueeze(dim = dimension),
                    dimension = dimension,
                    size =      n
                )
        
    def get_output_dimension(self,
        input_dimension:    int
    ) -> int:
        """# Get Output Dimension.
        
        Get output dimension after expansion (same as input dimension in this case).

        ## Args:
            * input_dimension   (int):  Input dimension.

        ## Returns:
            * int:  Output dimension, which in the case of the expander, is the same as the input.
        """
        # Simply return input.
        return input_dimension


"""# lucidium.spaces.continuous

Defines a continuous space for real-valued scalar actions or observations.
"""

__all__ = ["Continuous"]

from math                       import inf
from random                     import uniform
from typing                     import Tuple, Union

from lucidium.spaces.__base__   import Space

class Continuous(Space):
    """# Continuous (Space)

    Scalar-valued continuous space.
    """
    
    def __init__(self,
        lower:  Union[float, int] = -inf,
        upper:  Union[float, int] =  inf
    ):
        """# Instantiate Continuous (Space).

        ## Args:
            * lower (float):    Lower bound. Defaults to negative infinity.
            * upper (float):    Upper bound. Defaults to positive infinity.
        """
        # Define bounds.
        self._lower_:   Union[float, int] = float(lower)
        self._upper_:   Union[float, int] = float(upper)
        
        # Validate parameters.
        self.__post_init__()
        
    def __post_init__(self) -> None:
        """# Validate Parameters.
        
        ## Raises:
            * AssertionError:   If upper bound is less than lower bound.
        """
        assert self.upper > self.lower, \
            f"Upper bound {self.upper} must be greater than lower bound {self.lower}"
        
    # PROPERTIES ===================================================================================
    
    @property
    def bounds(self) -> Tuple[float, float]:
        """# (Continuous Space) Bounds

        Lower and upper bounds of space.
        """
        return self.lower, self.upper
    
    @property
    def lower(self) -> float:
        """# (Continuous Space) Lower Bound

        Lower bound of continuous space range.
        """
        return self._lower_
    
    @property
    def upper(self) -> float:
        """# (Continuous Space) Upper Bound

        Upper bound of continuous space range.
        """
        return self._upper_
    
    # METHODS ======================================================================================
    
    def contains(self,
        x:  Union[float, int]
    ) -> bool:
        """# (Continuous Space) Contains?

        ## Args:
            * x (Union[float, int]):    Scalar value being verified.

        ## Returns:
            * bool: True if x ∈ S.
        """
        return isinstance(x, (float, int)) and self.lower <= x < self.upper
    
    def sample(self) -> float:
        """# Sample (Continuous Space).
        
        Get a random number in the range [min, max) or [min, max] depending on rounding.

        ## Returns:
            * float:    Random scalar value from space.
        """
        return uniform(self.lower, self.upper)
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation.

        Object representation of continuous space.
        """
        return f"<Continuous(lower = {self.lower}, upper = {self.upper})>"


"""# lucidium.agents

This package defines various types of reinforcement learning agents.
"""

__all__ =   [
                # Abstract agent class.
                "Agent",
                
                # Concrete agent classes.
                "NeuralLogicMachine"
            ]

# Abstract agent class.
from lucidium.agents.__base__   import Agent

# Concrete agent classes.
from lucidium.agents.nlm        import NeuralLogicMachine


"""# lucidium.utilities.logger

Logging utilities.
"""

__all__ = ["get_child", "get_logger"]

from logging                import getLogger, Formatter, Logger, StreamHandler
from logging.handlers       import RotatingFileHandler
from os                     import makedirs
from sys                    import stdout

# Declare logger object.
LOGGER:         Logger

def get_logger(
    logger_name:    str,
    logging_level:  str =   "INFO",
    logging_path:   str =   "logs"
) -> Logger:
    """# Initialize Logger.
    
    Initialize a logger object with configuration provided.

    ## Args:
        * logger_name   (str):              Name attributed to logger object. This should usually 
                                            be the module identifier.
        * logging_level (str, optional):    Minimum logging level (DEBUG < INFO < WARNING < ERROR < 
                                            CRITICAL). Defaults to "INFO".
        * logging_path  (str, optional):    Path at which logs will be written. Defaults to "logs".

    ## Returns:
        * Logger:   Logger object, initialized with provided parameters.
    """
    # Ensure that logging path exists
    makedirs(name = logging_path, exist_ok = True)
    
    # Declare globals.
    global LOGGER

    # Initialize logger
    LOGGER =                                getLogger(
                                                name =          logger_name
                                            )

    # Set logging level
    LOGGER.setLevel(level = logging_level)

    # Define console handler
    stdout_handler: StreamHandler =         StreamHandler(
                                                stream =        stdout
                                            )

    # Define file handler
    file_handler:   RotatingFileHandler =   RotatingFileHandler(
                                                filename =      f"{logging_path}/{logger_name}.log",
                                                maxBytes =      1048576,
                                                backupCount =   10
                                            )
    
    # Define format for console handler.
    stdout_handler.setFormatter(Formatter("%(levelname)s | %(name)s | %(message)s"))
    
    # Define format fror file handler.
    file_handler.setFormatter(Formatter("%(asctime)s | %(levelname)s | %(name)s | %(message)s"))
    
    # Add handlers to logger.
    LOGGER.addHandler(hdlr = stdout_handler)
    LOGGER.addHandler(hdlr = file_handler)
    
    # Return logger object.
    return LOGGER
    
def get_child(
    logger_name:    str
) -> Logger:
    """# Initialize Child Logger.
    
    Declares a logger child descendent of root logger.

    ## Args:
        * logger_name   (str):  Name attributed to child logger.

    ## Returns:
        * Logger:   Child logger.
    """
    # Declare globals.
    global LOGGER
    
    # Simply return child logger.
    return LOGGER.getChild(logger_name)


"""# lucidium.spaces.multi_discrete

Defines a multi-discrete space (vector of discrete values).
"""

__all__ = ["MultiDiscrete"]

from random                     import randint
from typing                     import Tuple

from lucidium.spaces.__base__   import Space

class MultiDiscrete(Space):
    """# Multi-Discrete (Space)
    
    Vector-valued discrete space.
    """
    
    def __init__(self,
        shape:  Tuple[int, ...]
    ):
        """# Instantiate Multi-Discrete (Space).

        ## Args:
            * shape (Tuple[int, ...]):  Shape of space. Each dimension d will contain shape[d] 
                                        elements.
        """
        # Define shape.
        self._shape_:   Tuple[int, ...] =   shape
        
        # Validate parameters.
        self.__post_init__()
        
    def __post_init__(self) -> None:
        """# Validate Parameters.
        
        ## Raises:
            * AssertionError:   If any dimension of shape is not a positive integer.
        """
        return all(d > 0 for d in self.shape)
    
    # PROPERTIES ===================================================================================
    
    @property
    def shape(self) -> Tuple[int, ...]:
        """# (Multi-Discrete Space) Shape

        Represents the number of elements contained in each dimension of space.
        """
        return self._shape_
    
    # METHODS ======================================================================================
    
    def contains(self,
        v:  Tuple[int, ...]
    ) -> bool:
        """# (Multi-Discrete Space) Contains?

        ## Args:
            * v (Tuple[int, ...]):  Vector value being verified.

        ## Returns:
            * bool: True if v ∈ S.
        """
        return  all([
                    isinstance(v, tuple),
                    len(v) == len(self.shape),
                    all(
                        0 <= v_i < dimension
                        for v_i, dimension
                        in zip(v, self.shape)
                    )
                ])
        
    def sample(self) -> Tuple[int, ...]:
        """# Sample (Multi-Discrete Space).

        ## Returns:
            * Tuple[int, ...]:  Random vector value within space.
        """
        return (randint(0, dimension - 1) for dimension in self.shape)
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation.

        Object representation of multi-discrete space.
        """
        return f"<MultiDiscrete(shape = {self.shape})>"


"""# lucidium.registries.entry

Defines the structure and utility of a registration entry.
"""

from argparse   import _SubParsersAction
from typing     import Callable, List, Optional, Type

class RegistryEntry():
    """# Registry Entry
    
    Storage of registration entry data.
    """
    
    def __init__(self,
        cls:    Type,
        name:   str,
        tags:   Optional[List[str]] =   [],
        parser: Optional[Callable] =    None
    ):
        """# Instantiate Registration Entry

        ## Args:
            * cls       (Type):                 Class being registered.
            * name      (str):                  Name of entry.
            * tags      (Optional[List[str]]):  Tags that describe the taxonomy of the class being registered.
            * parser    (Optional[Callable]):   Argument parser handler.
        """
        # Define properties.
        self._cls_:     Type =      cls
        self._name_:    str =       name
        self._tags_:    List[str] = tags
        self._parser_:  Callable =  parser
        
    # PROPERTIES ===================================================================================
    
    @property
    def cls(self) -> Type:
        """# Registered Class"""
        return self._cls_
    
    @property
    def name(self) -> str:
        """# Registration Name."""
        return self._name_
    
    @property
    def parser(self) -> Optional[Callable]:
        """# Argument Parser Registration."""
        return self._parser_
    
    @property
    def tags(self) -> Optional[List[str]]:
        """# Entry Tags."""
        return self._tags_
        
    # METHODS ======================================================================================
    
    def contains_tag(self,
        tag:    str
    ) -> bool:
        """# (Entry) Contains Tag?

        ## Args:
            * tag   (str):  Tag being verified.

        ## Returns:
            * bool: True if entry contains tag.
        """
        return tag in self._tags_
    
    def register_parser(self,
        subparser: _SubParsersAction
    ) -> None:
        """# Register Parser.

        ## Args:
            * sub_parser    (_SubParsersAction):    Parent's sub parser.
        """
        self._parser_(subparser)
        
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation"""
        return f"""<{self._name_}({",".join(self._tags_)})>"""


"""# lucidium.spaces.box

Defines a multi-dimensional array space representation.
"""

__all__ = ["Box"]

from typing                     import Tuple, Union

from numpy                      import all, array, ndarray
from numpy.random               import uniform

from lucidium.spaces.__base__   import Space

class Box(Space):
    """# Box (Space)

    Multi-dimensional array space.
    """
    
    def __init__(self,
        lower:  Union[int, float],
        upper:  Union[int, float],
        shape:  Tuple[Union[int, float], ...],
        dtype:  Union[type, str] =              float
    ):
        """# Instantiate Box (Space).

        ## Args:
            * lower (Union[int, float]):                Lower bound.
            * upper (Union[int, float]):                Upper bound.
            * shape (Tuple[Union[int, float], ...]):    Shape of space.
            * dtype (Union[type, str], optional):       Data type (e.g., int, float). Defaults to 
                                                        float.
        """
        # Define bounds.
        self._lower_:   Union[int, float] =             lower
        self._upper_:   Union[int, float] =             upper
        
        # Define shape.
        self._shape_:   Tuple[Union[int, float], ...] = shape
        
        # Define data type.
        self._dtype_:   Union[type, str] =              dtype
    
    # PROPERTIES ===================================================================================
    
    @property
    def dtype(self) -> Union[type, str]:
        """# (Box Space) Data Type

        The data type of all elements within space.
        """
        return self._dtype_
    
    @property
    def lower_bound(self) -> Union[int, float]:
        """# (Box Space) Lower Bound

        Lower bound of all dimensions of space.
        """
        return self._lower_
    
    @property
    def shape(self) -> Tuple[Union[int, float], ...]:
        """# (Box Space) Shape.

        Shape of space.
        """
        return self._shape_
    
    @property
    def upper_bound(self) -> Union[int, float]:
        """(# Box Space) Upper Bound

        Upper bound of all dimensions of space.
        """
        return self._upper_
        
    # METHODS ======================================================================================
    
    def contains(self,
        x:  ndarray
    ) -> bool:
        """# (Box Space) Contains?

        ## Args:
            * x (ndarray):  Element being verified.

        ## Returns:
            * bool: True if x ∈ S.
        """
        # Convert x to an ndarray if it is another type.
        if not isinstance(x, ndarray): x: ndarray = array(x)
        
        # Evaluate existence.
        return  (
                    x.shape == self._shape_ and
                    all(x >= self._lower_)  and
                    all(x <= self._upper_)
                )
        
    def sample(self) -> ndarray:
        """# Sample (Box Space).

        ## Returns:
            * ndarray:  Random sample from box space.
        """
        return  uniform(
                    low =   self._lower_,
                    high =  self._upper_,
                    size =  self._shape_
                ).astype(self._dtype_)
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation.

        Object representation of box space.
        """
        return f"<Box(lower = {self._lower_}, upper = {self._upper_}, shape = {self._shape_})>"


"""# lucidium.spaces.composite

Defines a composite space combining named subspaces.
"""

__all__ = ["Composite"]

from typing                     import Any, Dict

from lucidium.spaces.__base__   import Space

class Composite(Space):
    """# Composite (Space)

    Dictionary-style combination of named subspaces.
    """
    
    def __init__(self,
        subspaces: Dict[str, Space]
    ):
        """# Instantiate Composite (Space).

        ## Args:
            * subspaces (Dict[str, Space]): Subspace mapping(s).
        """
        # Define subspaces.
        self._subspaces_:   Dict[str, Space] =  subspaces
        
    # PROPERTIES ===================================================================================
    
    @property
    def subspaces(self) -> Dict[str, Space]:
        """# (Composite Space) Spaces.

        ## Returns:
            * Dict[str, Space]: Mapping of subspaces.
        """
        return self._subspaces_
        
    # METHODS ======================================================================================
    
    def contains(self,
        m: Dict[str, Any]
    ) -> bool:
        """# (Composite Space) Contains?

        ## Args:
            * m (Dict[str, Any]):   Mapping of values being verified.

        ## Returns:
            * bool: True if all values of mapping exist within subspaces.
        """
        return  all([
                    isinstance(m, dict),
                    all(
                        [
                            key in self.subspaces,
                            self.subspaces[key.contains[m[key]]]
                        ]
                        for key in self.subspaces.items()
                    ) 
                ])
        
    def sample(self) -> Dict[str, Any]:
        """# Sample (Composite Space).

        ## Returns:
            * Dict[str, Any]:   Random value samples from each subspace.
        """
        return {key: space.sample() for key, space in self.subspaces.items()}
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation.

        Object representation of composite space.
        """
        return f"""<Composite(subspaces = {", ".join(f"{k}: {v}" for k, v in self.subspaces.items())})>"""


"""# lucidium.environments.base

This module provides the base class for all environments in the Lucidium framework.
"""

__all__ = ["Environment"]

from abc        import ABC, abstractmethod
from typing     import Any, Dict, Tuple

class Environment(ABC):
    """Abstract Environment Class
    
    This class defines the base structure for an environment that can be interacted with by agents.
    
    ## Properties:
        * action_space  (Any):  Actions possible within the environment.
        * state_space   (Any):  Quantification of possible states in which the environment can be 
                                observed.
    """
        
    # PROPERTIES ===================================================================================
    
    @property
    @abstractmethod
    def action_space(self) -> Any:
        """# Action Space (Any)
        
        This property should return the actions possible within the environment.
        """
        pass
    
    @property
    @abstractmethod
    def state_space(self) -> Any:
        """# State Space (Any)
        
        This property should return the quantification of possible states in which the environment 
        can be observed.
        """
        pass
    
    # METHODS ======================================================================================
    
    @abstractmethod
    def reset(self) -> Any:
        """# Reset Environment.
        
        This method should reset the environment to its initial state and return the initial state.
        
        ## Returns:
            * Any:  Initial state of the environment.
        """
        pass
    
    @abstractmethod
    def step(self,
        action: Any
    ) -> Tuple[Any, float, bool, Dict]:
        """# Step.

        The second critical step in the reinforcement learning loop.

        Applies the agent's action, updates the environment state, and returns the resulting new 
        state, reward, done flag, and optional metadata.
        
        ## Args:
            * action    (Any):  Action to be taken in the environment.
        
        ## Returns:
            * Tuple[Any, float, bool, Dict]:
                - Any:      Next state of the environment
                - float:    Reward received
                - bool:     Done flag indicating if the episode has ended
                - Dict:     Additional information
        """
        pass


"""# lucidium.args

Lucidium application argument definitions and parsing.
"""

__all__ = ["parse_lucidium_arguments"]

from argparse                           import ArgumentParser, _ArgumentGroup, Namespace, _SubParsersAction

def parse_lucidium_arguments() -> Namespace:
    """# Parse Lucidium Arguments.
    
    This function should be called at the entry point of the Lucidium application. The name space of 
    argument keys and values that it provides will be passed/provided to subsequent module entry 
    points.
    
    ## Returns:
        * NameSpace:    Name space of parsed arguments and their values.
    """
    # Initialize primary parser
    _parser_:       ArgumentParser =    ArgumentParser(
        prog =          "lucidium",
        description =   """Suite of environments, models, & methods in pursuit of achieving organic 
                        reasoning & logic by means of deep reinforcement learning."""
    )

    # Initialize sub-parser
    _subparser_:    _SubParsersAction = _parser_.add_subparsers(
        dest =          "command",
        help =          "Lucidium commands."
    )

    # +============================================================================================+
    # | BEGIN ARGUMENTS                                                                            |
    # +============================================================================================+

    # LOGGING ======================================================================================
    _logging_:      _ArgumentGroup =    _parser_.add_argument_group(
        title =         "Logging",
        description =   "Logging configuration."    
    )

    _logging_.add_argument(
        "--logging-level",
        type =          str,
        choices =       ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
        default =       "INFO",
        help =          """Minimum logging level (DEBUG < INFO < WARNING < ERROR < CRITICAL). 
                        Defaults to "INFO"."""
    )

    _logging_.add_argument(
        "--logging-path",
        type =          str,
        default =       "logs",
        help =          """Path at which logs will be written. Defaults to "./logs/"."""
    )

    # +============================================================================================+
    # | END ARGUMENTS                                                                              |
    # +============================================================================================+

    from lucidium.registries import AGENT_REGISTRY, ENVIRONMENT_REGISTRY
    
    # Force loading of all agents and environments.
    AGENT_REGISTRY.load_all()
    ENVIRONMENT_REGISTRY.load_all()
    
    # Register agent and environment parsers.
    AGENT_REGISTRY.register_parsers(parent_subparser = _subparser_)
    ENVIRONMENT_REGISTRY.register_parsers(parent_subparser = _subparser_)

    # Parse arguments
    return _parser_.parse_args()


"""# lucidium.environments.tic_tac_toe.base

Defines the Tic Tac Toe game framework.
"""

from logging                                        import Logger
from typing                                         import Any, Dict, List, override, Tuple, Union

from torch                                          import Tensor

from lucidium.environments.tic_tac_toe.__args__     import register_tic_tac_toe_parser
from lucidium.environments.tic_tac_toe.components   import *
from lucidium.environments.__base__                 import Environment
from lucidium.registries                            import register_environment
from lucidium.spaces                                import Box, Discrete
from lucidium.symbolic                              import Predicate
from lucidium.utilities                             import get_child

@register_environment(
    name =      "tic-tac-toe",
    tags =      ["game", "two-player", "turn-based", "discrete", "strategic"],
    parser =    register_tic_tac_toe_parser
)
class TicTacToe(Environment):
    """# Tic-Tac-Toe (Environment)
    
    Tic-Tac-Toe is a two-player game where players take turns marking a square in a 3x3 grid. The 
    first player to align three of their marks horizontally, vertically, or diagonally wins the 
    game.
    """
    
    def __init__(self,
        size:                   int =    3,
        win_reward:             float =  1.0,
        loss_penalty:           float = -1.0,
        draw_reward:            float =  0.0,
        invalid_move_penalty:   float = -0.5,
        move_penalty:           float = -0.1,
        **kwargs
    ):
        """# Instantiate Tic-Tac-Toe Environment

        ## Args:
            * size                  (int, optional):    Size of (square) board. Defaults to 3.
            * win_reward            (float, optional):  Reward yielded from winning a game. Defaults 
                                                        to 1.0.
            * loss_penalty          (float, optional):  Penalty incurred by losing a game. Defaults 
                                                        to -1.0.
            * draw_reward           (float, optional):  Reward yielded by ending game in a draw 
                                                        (neither player wins). Defaults to 0.0.
            * invalid_move_penalty  (float, optional):  Penalty incurred for attempting an invalid 
                                                        move. Defaults to -0.5.
            * move_penalty          (float, optional):  Cost of making a single move. Defaults to 
                                                        -0.1.
        """
        # Inititialize environment.
        super(TicTacToe, self).__init__(**kwargs)
        
        # Initialize logger.
        self.__logger__:                Logger =    get_child("tic-tac-toe")
        
        # Define environment properties
        self._size_:                    int =       size
        self._win_reward_:              float =     win_reward
        self._loss_penalty_:            float =     loss_penalty
        self._draw_reward_:             float =     draw_reward
        self._invalid_move_penalty_:    float =     invalid_move_penalty
        self._move_penalty_:            float =     move_penalty
        
        # Instantiate board.
        self._board_:                   Board =     Board(size = size)
        
        # Define players.
        self._player_x_:                Player =    Player.X
        self._player_o_:                Player =    Player.O
        self._current_player_:          Player =    self._player_x_
        self._winner_:                  Player =    Player.EMPTY
        
        # Log for debugging.
        self.__logger__.debug(f"Initialized TicTacToe environment ({locals()})")
        
    # PROPERTIES ===================================================================================
    
    @override
    @property
    def action_space(self) -> Discrete:
        """# (Tic-Tac-Toe) Action Space"""
        return Discrete(n = self._size_ ** 2)
    
    @override
    @property
    def state_space(self) -> Box:
        """# (Tic-Tac-Toe) Observation Space"""
        return Box(lower = -1, upper = 1, shape = (self._size_, self._size_), dtype = int)
    
    # METHODS ======================================================================================
    
    @override
    def reset(self) -> Union[Tensor, List[Predicate]]:
        """# Reset (Environment).
        
        Reset environment to initial state.

        ## Returns:
            * Union[Tensor, List[Predicate]]:   Observation of environment state after reset.
        """
        # Reset board.
        self._board_.reset()
        
        # Provide observation.
        return self.observe()
    
    @override
    def step(self,
        action: int
    ) -> Tuple[Union[Tensor, List[Predicate]], float, bool, Dict[str, Any]]:
        """# Step
        
        Update environment based on action submitted.

        ## Args:
            * action    (int):  Action submitted by agent.

        ## Returns:
            * Tuple[Union[Tensor, List[Predicate]], float, bool, Dict[str, Any]]:
                * Updated state after action
                * Reward yielded/penalty incurred by action
                * True if environment has reached a terminal state
                * Metadata
        """
        # If game has already concluded, an action cannot be submitted.
        if self._done_: raise RuntimeError(f"Game has already concluded. Call reset() to start a new game.")
        
        # Log action for debugging.
        self.__logger__.debug(f"Player {self._current_player_} submitted action: {action}")
        
        # If action is not valid...
        if  not self.action_space.contains(action) or \
            not self._board_.mark_by_index(index = action, entry = self._current_player_.symbol):
                
            # Assign penalty.
            return  self._get_observation_(),   \
                    self._invalid_move_penalty_,\
                    False,                      \
                    {"event": f"Player {self._current_player_} attempted invalid move {action}"}
                    
        # If player made a winning move...
        if self._board_.has_winner:
            
            # Assign reward.
            return  self._get_observation_(),   \
                    self._win_reward_,          \
                    True,                       \
                    {"event": f"Player {self._current_player_} won the game."}
                    
        # If board is full...
        if self._board_.is_full:
            
            # Assign penalty.
            return  self._get_observation_(),   \
                    self._draw_reward_,         \
                    True,                       \
                    {"event": f"Game ended in draw."}
                    
        # Otherwise, simply assign cost of move.
        return  self._get_observation_(),   \
                self._move_penalty_,        \
                False,                      \
                {"event": f"Player {self._current_player_} made move {action}"}
                
    # HELPERS ======================================================================================
    
    def _get_observation_(self) -> Dict[str, Union[Tensor, List[Predicate]]]:
        """# Get Observation.
        
        Provide observation of environment's current state.

        ## Returns:
            * Dict[str, Union[Tensor, List[Predicate]]]:
                * Tensor state representation
                * Symbolic predicates that apply to current state
        """
        
        
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation"""
        return f"""TicTacToe(size = {self._size_})"""
    
    def __str__(self) -> str:
        """# String Representation"""
        return str(self._board_)


"""# lucidium.version

Version tracking utility.
"""

__author__ = "Gabriel C. Trahan"
__author_email__ = "gabrieltrahan777@hotmail.com"
__description__ = """Experiments in neuro-symbolic reinforcement learning in the pursuit of developing agents/methods to aid in organic reasoning, inference, and logic."""
__title__ = "lucidium"
__url__ = "https://github.com/theokoles7/lucidium"
__version__ = "0.1.2"        
__version_info__ = tuple(int(i) for i in __version__.split("."))


"""# lucidium.spaces.discrete

Defines a discrete (integer-indexed) space.
"""

__all__ = ["Discrete"]

from random                     import randint

from lucidium.spaces.__base__   import Space

class Discrete(Space):
    """# Discrete (Space)

    Scalar-valued discrete space.
    """
    
    def __init__(self,
        n:  int
    ):
        """# Instantiate Discrete (Space).

        ## Args:
            * n (int):  Number of elements.
        """
        # Define number of elements.
        self._n_:   int =   n
        
        # Validate parameters.
        self.__post_init__()
        
    def __post_init__(self) -> None:
        """# Validate parameters.
        
        ## Raises:
            * AssertionError:   If n is not a positive integer.
        """
        assert self.n > 0, f"Discrete space must have positive number of elements, got n = {self.n}"
    
    # PROPERTIES ===================================================================================
    
    @property
    def n(self) -> int:
        """# (Discrete Space) N

        Number of elements in space.
        """
        return self._n_
        
    # METHODS ======================================================================================
    
    def contains(self,
        x:  int
    ) -> bool:
        """# (Discrete Space) Contains?
        
        ## Args:
            * x (int):  Value being verified.

        ## Returns:
            * bool: True if x ∈ S.
        """
        return isinstance(x, int) and 0 <= x < self.n
    
    def sample(self) -> int:
        """# Sample (Discrete Space).

        ## Returns:
            * int:  Random element from space.
        """
        return randint(0, self.n - 1)
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation.

        Object representation of discrete space.
        """
        return f"<Discrete(n = {self.n})>"


"""# lucidium.environments.tic_tac_toe.components.cell

Implementation of cell component for tic-tac-toe board.
"""

from typing                                                 import Tuple, Union

from torch                                                  import Tensor

from lucidium.environments.tic_tac_toe.components.players   import Player
from lucidium.symbolic                                      import Predicate

class Cell():
    """(Tic-Tac-Toe) Cell

    Represents a single square on the Tic-Tac-Toe board.
    Tracks the occupying Player and supports symbolic and tensor interfaces.
    """
    
    def __init__(self,
        coordinate: Tuple[int, int]
    ):
        """# Instantiate Cell.

        ## Args:
            * coordinate    (Tuple[int, int]):  Row, column coordinate at which cell is located on 
                                                board.
        """
        # Define location.
        self._coordinate_:  Tuple[int, int] =   coordinate
        self._row_:         int =               coordinate[0]
        self._column_:      int =               coordinate[1]
        
        # Start with empty cell.
        self._player_:      Player =            Player.EMPTY
        
    # PROPERTIES ===================================================================================
    
    @property
    def entry(self) -> str:
        """# (Cell) Entry

        Current entry in cell.
        """
        return self._player_.symbol
    
    @property
    def is_empty(self) -> bool:
        """# (Cell) is Empty?

        True if cell is empty.
        """
        return self._player_ == Player.EMPTY
    
    # METHODS ======================================================================================
    
    def mark(self,
        entry:  Union[str, int, Tensor]
    ) -> bool:
        """# Mark (Cell).

        ## Args:
            * entry (Union[str, int, Tensor]): Player entry representation:
                * Symbol ("X", "O", " ")
                * Number (1, -1, 0)
                * One-hot tensor (e.g., [1, 0, 0])
                
        ## Returns:
            * bool: True if cell was marked.
        """
        # If cell is not empty, it cannot be marked.
        if not self.is_empty: return False
        
        # Otherwise, mark cell.
        self._set_player_(player = entry)
        
        # Indicate that cell was marked.
        return True
    
    def reset(self) -> None:
        """# Reset (Cell).

        Reset cell to initial state.
        """
        self._player_: Player = Player.EMPTY
        
    def to_symbolic(self) -> Predicate:
        """# (Cell) to Symbolic Representation.

        ## Returns:
            * Predicate: Predicate representation of cell.
        """
        return Predicate("position", self._player_.symbol, self._row_, self._column_)
    
    def to_tensor(self) -> Tensor:
        """# (Cell) to Tensor
        
        ## Returns:
            * Tensor:   One-hot tensor encoding of player.
        """
        return self._player_.encoding
        
    # HELPERS ======================================================================================
    
    def _set_player_(self,
        player: Union[Player, str, int, Tensor]
    ) -> None:
        """# Set Player.

        ## Args:
            * entry (Union[Player, str, int, Tensor]): Player entry representation:
                * Player object
                * Symbol ("X", "O", " ")
                * Number (1, -1, 0)
                * One-hot tensor (e.g., [1, 0, 0])
        """
        # Match data type.
        match player:
            
            # Player object.
            case Player()   as p:   self._player_ = p
            
            # Symbol.
            case str()      as p:   self._player_ = Player.from_symbol(symbol = p)
            
            # Number.
            case int()      as p:   self._player_ = Player.from_number(number = p)
            
            # One-hot tensor.
            case Tensor()   as p:   self._player_ = Player.from_encoding(encoding = p)
            
            # Invalid representation.
            case _: raise TypeError(f"Player representation type not recognized: {player} ({type(player)})")
            
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation"""
        return f"""<Cell(row = {self._row_}, column = {self._column_}, player = {self._player_})>"""
    
    def __str__(self) -> str:
        """# String Representation"""
        return str(self._player_)


"""# lucidium..symbolic

This package defines various structures needed for symbolic learning.
"""

__all__ =   [
                "Predicate"
            ]

from lucidium.symbolic.predicate    import Predicate


"""# lucidium

Experiments in neuro-symbolic reinforcement learning in the pursuit of developing agents/methods to 
aid in organic reasoning, inference, and logic.
"""

__all__ =   [
                "__author__",
                "__author_email__",
                "__description__",
                "__title__",
                "__url__",
                "__version__",
                "__version_info__",
            ]

from lucidium._version_ import  (
                                    __author__,
                                    __author_email__,
                                    __description__,
                                    __title__,
                                    __url__,
                                    __version__,
                                    __version_info__,
                                )


"""# lucidium.environments.tic_tac_toe.components.glyphs

This module provides the players for the Tic Tac Toe environment.
"""

__all__ = ["Player"]

from enum           import Enum

from torch          import equal, tensor, Tensor

class Player(Enum):
    """(Tic-Tac-Toe) Player

    This enumeration defines the players used to represent the Tic Tac Toe environment.
    """
    
    EMPTY = {"number":  0, "symbol": " ", "encoding": tensor((1, 0, 0))}
    X     = {"number":  1, "symbol": "X", "encoding": tensor((0, 1, 0))}
    O     = {"number": -1, "symbol": "O", "encoding": tensor((0, 0, 1))}
    
    # PROPERTIES ===================================================================================
    
    @property
    def name(self) -> str:
        """# (Player) Name

        Symbolic representation of the player (e.g., 'X', 'O').
        """
        return self.value["symbol"]
    
    @property
    def number(self) -> int:
        """# (Player) Number

        Integer value of the player: +1 (X), -1 (O), or 0 (empty).
        """
        return self.value["number"]
    
    @property
    def encoding(self) -> Tensor:
        """# (Player) One-Hot Encoding.

        One-hot encoding representation of player.
        """
        return self.value["encoding"]
    
    @property
    def opponent(self) -> "Player":
        """# Opponent

        Respective opponent of player.
        """
        # If I am X, my opponent is O.
        if self == Player.X: return Player.O
        
        # If I am O, my opponent is X.
        if self == Player.O: return Player.X
        
        # Otherwise, empty.
        return Player.EMPTY
    
    @property
    def symbol(self) -> str:
        """# (Player) Symbol

        Symbolic representation of the player (e.g., 'X', 'O').
        """
        return self.value["symbol"]
    
    # CLASS METHODS ================================================================================
    
    @classmethod
    def from_number(cls,
        number:  int
    ) -> "Player":
        """# (Player) from Value

        Return the player corresponding to the given numeric value.

        ## Args:
            * number    (int):  +1 for player or -1 for opponent.

        ## Raises:
            * ValueError:   If player number is not known.

        ## Returns:
            * Player:   Player initialized from value.
        """
        # For each defined player...
        for player in cls:
            
            # If it matches the value provided, return that player.
            if player.number == number: return player
            
        # Otherwise, report invalid player number.
        raise ValueError(f"No player found with number {number}")
    
    @classmethod
    def from_encoding(cls,
        encoding: Tensor
    ) -> "Player":
        """# (Player) from One-Hot Encoding.

        ## Args:
            * encoding  (Tensor):  One-hot encoding representation of player.

        ## Raises:
            * ValueError:   If player encoding is not known.

        ## Returns:
            * Player:   Player initialized from one-hot encoding.
        """
        # For each defined player...
        for player in cls:
            
            # If it matches the value provided, return that player.
            if equal(player.encoding, encoding): return player
            
        # Otherwise, report invalid player number.
        raise ValueError(f"No player found with encoding {encoding}")
    
    @classmethod
    def from_symbol(cls,
        symbol: str
    ) -> "Player":
        """# (Player) from Symbol

        Return the player corresponding to the given symbol value.

        ## Args:
            * symbol    (str):  "X" for player or "O" for opponent.

        ## Raises:
            * ValueError:   If player symbol is not known.

        ## Returns:
            * Player:   Player initialized from symbol.
        """
        # For each defined player...
        for player in cls:
            
            # If it matches the value provided, return that player.
            if player.symbol == symbol.upper(): return player
            
        # OTherwise, report invalid player number.
        raise ValueError(f"No player found with symbol {symbol}")
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation.
        
        Return the object representation of the glyph.
        """
        return f"<{self.name}>"
    
    def __str__(self) -> str:
        """# String Representation.
        
        Return the string representation of the glyph.
        """
        return self.symbol


"""# lucidium.agents.nlm.layers.reduction

Defines the dimension reduction layer for neural logic machines.
"""

__all__ = ["DimensionReducer"]

from torch                          import max, min, Size, stack, Tensor
from torch.nn                       import Module

from lucidium.agents.nlm.__util__   import exclude_mask, mask_value

class DimensionReducer(Module):
    """# Dimension Reducer

    Reduces a variable via quantifiers (exists/forall), implemented by max/min pooling.
    
    Adapted from : https://github.com/google/neural-logic-machines/blob/master/difflogic/nn/neural_logic/modules/dimension.py
    """
    
    def __init__(self,
        dimension:      int,
        exclude_self:   bool =  True,
        exists:         bool =  True
    ):
        """# Initialize Dimension Reducer.

        ## Args:
            * dimension     (int):  Dimension along which reduction will occur.
            * exclude_self  (bool): Exclude the self variable in the reduction.
            * exists        (bool): Use "exists" (True) or "forall" (False) quantifier.
        """
        # Initialize module.
        super(DimensionReducer, self).__init__()
        
        # Define dimension.
        self._dimension_:       int =   dimension
        
        # Define flags.
        self._exclude_self_:    bool =  exclude_self
        self._exists_:          bool =  exists
        
    # PROPERTIES ===================================================================================
    
    @property
    def dimension(self) -> int:
        """# Dimension (int)
        
        Dimension along which reduction will occur.
        """
        return self._dimension_
    
    @property
    def exclude_self(self) -> bool:
        """# Exclude Self (bool)
        
        Whether to exclude the self variable in the reduction.
        """
        return self._exclude_self_
    
    @property
    def exists(self) -> bool:
        """# Exists (bool)
        
        Whether to use "exists" (True) or "forall" (False) quantifier.
        """
        return self._exists_
    
    # METHODS ======================================================================================
        
    def forward(self,
        inputs: Tensor,
        n:      int | None =    None
    ) -> Tensor:
        """# Reduce Input.
        
        Reduce input tensor by performing pooling operations based on the quantifier type.

        ## Args:
            * inputs    (Tensor):   Input tensor being reduced.

        ## Returns:
            * Tensor:   Reduced tensor.
        """
        # Reference size of input tensor.
        shape:      Size =          inputs.size()
        
        # Create two copies of input for pooling.
        input1:     Tensor =        inputs
        input2:     Tensor =        inputs
        
        # If excluding self...
        if self.exclude_self:
            
            # Calculate mask.
            mask:   Tensor =    exclude_mask(
                                    inputs =    inputs,
                                    count =     self.dimension,
                                    dimension = -1 - self.dimension
                                )
            
            # Mask input with zeros for excluded values.
            input1: Tensor =    mask_value(inputs = inputs, mask = mask, value = 0.0)
            
            # Mask input with ones for remaining values.
            input2: Tensor =    mask_value(inputs = inputs, mask = mask, value = 1.0)
            
        # If using "exists" quantifier...
        if self.exists:
            
            # Adjust shape to accommodate both exists and forall values.
            shape:  Size =      shape[:-2] + (shape[-1] * 2,)
            
            # Apply max pooling along the specified dimension.
            exists: Tensor =    max(input = input1, dim = -2)[0]
            
            # Apply min pooling along the specified dimension.
            forall: Tensor =    min(input = input2, dim = -2)[0]
            
            # Stack and reshape the results.
            return stack(tensors = (exists, forall), dim = -1).view(shape)
        
    def get_output_dimension(self,
        input_dimension:    int
    ) -> int:
        """# Get Output Dimension.
        
        Get output dimension after reduction.

        ## Args:
            * input_dimension   (int):  Input dimension.

        ## Returns:
            * int:  Output dimension after reduction.
        """
        # If using the "exists" quantifier, output dimension is doubled.
        if self.exists: return input_dimension * 2
        
        # Otherwise, simply reutrn input dimension.
        return input_dimension


"""# lucidium.agents.nlm.layers.linear

Defines a linear layer with activation for neural logic machines.
"""

__all__ = ["LinearLayer"]

from torch.nn   import Linear, ReLU, Sequential

class LinearLayer(Sequential):
    """# Linear Layer
    
    A fully connected layer with activation.

    Adapted from jatorch.nn.cnn.layers: https://github.com/vacancy/Jacinle/blob/master/jactorch/nn/cnn/layers.py
    """
    
    def __init__(self,
        in_features:    int,
        out_features:   int
    ):
        """# Initialize Linear Layer.
        
        ## Args:
            * in_features   (int):  The number of input features.
            * out_features  (int):  The number of output features.
        """
        # Initialize sequential module.
        super(LinearLayer, self).__init__([
            Linear(
                    in_features =   in_features,
                    out_features =  out_features,
                    bias =          True
            ),
            ReLU(
                    inplace =       True
            )
        ])
        
        # Define features.
        self._in_features_:     int =   in_features
        self._out_features_:    int =   out_features
        
    # PROPERTIES ===================================================================================
        
    @property
    def input_dimension(self) -> int:
        """# Input Dimension (int)
        
        Dimension of input features.
        """
        return self._in_features_
    
    @property
    def output_dimension(self) -> int:
        """# Output DImension (int)
        
        Dimension of output features.
        """
        return self._out_features_
    
    # METHODS ======================================================================================
    
    def reset_parameters(self) -> None:
        """# Reset Parameters.
        
        Reset parameters of linear layer.
        """
        # For each module in layer...
        for module in self.modules():
            
            # If it is a linear layer, reset the parameters.
            if isinstance(module, Linear): module.reset_parameters()


"""# lucidium.spaces.multi_continuous

Defines a multi-dimensional continuous space.
"""

__all__ = ["MultiContinuous"]

from random                     import uniform
from typing                     import Tuple, Union

from lucidium.spaces.__base__   import Space

class MultiContinuous(Space):
    """# Multi-Continuous (Space)

    Vector-valued continuous space.
    """
    
    def __init__(self,
        bounds: Tuple[Tuple[Union[float, int], Union[float, int]]]
    ):
        # Define bounds.
        self._bounds_:  Tuple[Tuple[Union[float, int], Union[float, int]]] =    bounds
        
    def __post_init__(self) -> None:
        """# Validate Parameters.
        
        ## Raises:
            * AssertionError:   If any dimension's upper bound is less than its corresponding lower 
                                bound.
        """
        # For each dimension...
        for d, dimension in enumerate(self.bounds, start = 1):
            
            # Assert that upper bound is greater than lower bound.
            assert dimension[0] < dimension[1], \
                f"Dimension {d}: Upper bound {dimension[1]} must be greater than lower bound {dimension[0]}"
        
    # PROPERTIES ===================================================================================
    
    @property
    def bounds(self) -> Tuple[Tuple[float, float]]:
        """# (Multi-Continuous Space) Bounds.

        ## Returns:
            * Tuple[Tuple[float, ...]]: Multi-dimensional bounds of space.
        """
        return self._bounds_
    
    # METHODS ======================================================================================
    
    def contains(self,
        v:  Tuple[Union[float, int], ...]
    ) -> bool:
        """# (Multi-Continuous Space) Contains?

        ## Args:
            * v (Tuple[Union[float, int], ...]):    Vector value being verified.

        ## Returns:
            * bool: True if v ∈ S.
        """
        return  all([
                    isinstance(v, tuple),
                    len(v) == len(self.bounds),
                    all(
                        bound[0] <= v_i < bound[1]
                        for v_i, bound
                        in zip(v, self.bounds)
                    )
                ])
        
    def sample(self) -> Tuple[float, ...]:
        """# Sample (Multi-Continuous Space).

        ## Returns:
            * Tuple[float, ...]:    Random vector value from space.
        """
        return (uniform(bound[0], bound[1]) for bound in self.bounds)
    
    # DUNDERS ======================================================================================
    
    def __repr__(self) -> str:
        """# Object Representation.

        Object representation of multi-continuous space.
        """
        return f"<MultiContinuous(bounds = {self.bounds})>"


"""# lucidium.spaces.base

Defines the abstract class representation for action and observation spaces.
"""

__all__ = ["Space"]

from abc    import ABC, abstractmethod
from typing import Any

class Space(ABC):
    """# Abstract Space

    Defines the abstract interface for an action or observation space.
    """
    
    # METHODS ======================================================================================
    
    @abstractmethod
    def contains(self,
        value:  Any
    ) -> bool:
        """(Space) Contains?
        
        ## Args:
            * value (Any):  Value being verified.

        True if value is an element of space.
        """
        pass
    
    @abstractmethod
    def sample(self) -> Any:
        """# Sample (Space).

        Randomly sample a valid element from space.
        """
        pass
    
    # DUNDERS ======================================================================================
    
    @abstractmethod
    def __repr__(self) -> str:
        """# Object Representation.

        Object representation of space.
        """
        pass


"""# lucidium.setup

# Lucidium setup utility.
"""

from pathlib    import Path
from setuptools import find_packages, setup
from typing     import Any, Dict

def get_long_description():
    """# Get Long Description.
    
    Relay long description from README file.
    """
    # Open file.
    with open(Path(__file__).parent / "README.md", encoding = "utf-8") as f:
        
        # Provide content.
        return f.read()

def get_version() -> str:
    """# Get Version.
    
    Fetch version from internal version file.

    ## Returns:
        * str:  Current version.
    """
    # Initialize dictionary to store globals.
    version_vars:   Dict[str, Any] =    {}
    
    # Open file.
    with open(Path(__file__).parent / "lucidium" / "_version_.py") as f:
        
        # Read data into dictionary.
        exec(f.read(), version_vars)
        
    # Provide version.
    return version_vars["__version__"]

# Set up package.
setup(
    name =                          "lucidium",
    version =                       get_version(),
    author =                        "Gabriel C. Trahan",
    author_email =                  "gabrieltrahan777@hotmail.com",
    description =                   """Experiments in neuro-symbolic reinforcement learning in the 
                                    pursuit of developing agents/methods to aid in organic 
                                    reasoning, inference, and logic.""",
    long_description =              get_long_description(),
    long_description_content_type = "text/markdown",
    license =                       "GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007",
    license_files =                 ("LICENSE"),
    url =                           "https://github.com/theokoles7/lucidium",
    packages =                      find_packages(),
    python_requires =               ">=3.10",
    install_requires =              [
                            
                                    ],
    entry_points =                  {
                                        "console_scripts":  [
                                                                "lucidium=lucidium.__main__:main",
                                                            ],
                                    },
    classifiers =                   [
                                        "Intended Audience :: Developers",
                                        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                                        "Operating System :: OS Independent",
                                        "Programming Language :: Python :: 3",
                                        "Programming Language :: Python :: 3.10",
                                        "Programming Language :: Python :: 3.11",
                                        "Programming Language :: Python :: 3.12",
                                        "Topic :: Scientific/Engineering :: Artificial Intelligence",
                                    ],
)


"""# lucidium.agents.nlm.util

Defines utility functions for the Neural Logic Machine (NLM) package.
"""

__all__ =   [
                "broadcast",
                "concatenate_shape",
                "exclude_mask",
                "mask_value",
            ]

from collections.abc    import Sequence
from typing             import List, Tuple, Union

from torch              import arange, long, Size, Tensor
        
def broadcast(
    tensor:     Tensor,
    dimension:  int,
    size:       int
) -> Tensor:
    """# Broadcast Tensor.
    
    Broadcast a specific dimension for `size` times. Originally the dimension size must be 1.
    
    Adapted from jatorch.functional.shape: https://github.com/vacancy/Jacinle/blob/master/jactorch/functional/shape.py

    ## Args:
        * tensor    (Tensor):   Tensor being broadcasted.
        * dimension (int):      Dimension of tensor being broadcasted.
        * size      (int):      Size of target dimension.

    ## Returns:
        * Tensor:   Broadcasted tensor.
    """
    # If dimension is less than zero, increment.
    if dimension < 0: dimension += tensor.dim()
    
    # Assert that the tensor's dimension size is one.
    assert tensor.size(dim = dimension) == 1
    
    # Broadcast tensor.
    return  tensor.expand(
                size =  concatenate_shape(
                            tensor.size()[:dimension],
                            size,
                            tensor.size()[dimension + 1:]
                        )
            )
    

def concatenate_shape(
    *shapes:    Union[Size, Tuple[int, ...], List[int], int]
) -> Tuple[int, ...]:
    """# Concatenate Shapes into Tuple.
    
    Concatenate shapes into a tuple. The values can be either torch.Size, tuple, list, or integer.
    
    Adapted from jatorch.functional.shape: https://github.com/vacancy/Jacinle/blob/master/jactorch/functional/shape.py
    
    ## Returns:
        * Tuple[int, ...]:  Tuple of shapes.
    """
    # Initialize output list.
    output: List =  []
    
    # For shape provided...
    for s in shapes:
        
        # If shape is a sequence, add elements to output list.
        if isinstance(s, Sequence): output.extend(s)
        
        # Otherwise simply append single element.
        output.append(int(s))
        
    # Provide tuple of shapes.
    return tuple(output)


def exclude_mask(
    inputs:     Tensor,
    count:      int =   2,
    dimension:  int =   1
) -> Tensor:
    """# Produce Exclusive Mask.
    
    Produce an exclusive mask for a tensor, ensuring mutually exclusive coordinates.

    Specifically, for `cnt=2`, given an array `a[i, j]` of size `n * n`, this function produces a 
    mask where only `a[i, j] == 1` if and only if `i != j`.
    
    Adapted from: https://github.com/google/neural-logic-machines/blob/master/difflogic/nn/neural_logic/modules/_utils.py

    ## Args:
        * inputs    (Tensor):           The tensor to be masked.
        * count     (int, optional):    The number of dimensions over which the operation is 
                                        performed. Default is 2.
        * dimension (int, optional):    The starting dimension for the exclusive mask. Default is 1.

    ## Returns:
        * Tensor:   A mask tensor where each coordinate pair is mutually exclusive.
    """
    # Assert that the count is greater than zero.
    assert count > 0, f"Count({count}) expected to be greater than zero."
    
    # Adjust dimension if negative.
    if dimension < 0: dimension += inputs.dim()
    
    # Reference size of input.
    n:      Size =      inputs.size(dim = dimension)
    
    # Assert that size of all input dimensions are equal.
    for i in range(1, count): assert n == inputs.size(dim = dimension + i), \
        f"Size of dimensions in count must be equal."
        
    # Initialize range of values along dimension.
    values: Tensor =    arange(start = 0, end = n, dtype = long, device = inputs.device)
    
    # Initialize list to track index ranges.
    q:      List =      []
    
    # For each count...
    for i in range(count):
        
        # Start with range of values 0 to n-1.
        p:  Tensor =    values
        
        # For other axes...
        for j in range(count):
            
            # Make sure we're not looking at the same axis.
            if i != j:
                
                # Unsqueeze the indices along the specified dimension.
                p:  Tensor =    p.unsqueeze(j)
                
        # Expand the indices to cover the entire grid.
        p:  Tensor =    p.expand(size = (n,) * count)
        
        # Append to expanded indices list.
        q.append(p)
        
    # Initialize mask where first condition holds.
    mask:   Tensor =    q[0] == q[0]
    
    # Apply mutually exclusive condition.
    for i in range(count):
        for j in range(count):
            if i != j: mask *= q[i] != q[j]
    
    # Unsqueeze mask along leading dimensions.
    for i in range(dimension):                          mask.unsqueeze_(dim = 0)
    for j in range(inputs.dim() - dimension - count):   mask.unsqueeze_(dim = -1)
    
    # Expand mask to match the input size.
    return mask.expand(size = inputs.size()).float()
    
def mask_value(
    inputs: Tensor,
    mask:   Tensor,
    value:  float
) -> Tensor:
    """# Apply Mask to Tensor.
    
    Apply a mask to a tensor, setting the masked elements to a specified value.

    This function will multiply the `inputs` tensor with the `mask`, and set all elements where the 
    `mask` is zero to the given `value`.
    
    Adapted from: https://github.com/google/neural-logic-machines/blob/master/difflogic/nn/neural_logic/modules/_utils.py

    ## Args:
        * inputs    (Tensor):   The input tensor to be modified.
        * mask      (Tensor):   A binary mask tensor, where `1` means the value should remain, and 
                                `0` means it should be replaced.
        * value     (float):    The value to replace the elements where the mask is `0`.

    ## Returns:
        * Tensor:   The modified tensor after applying the mask.
    """
    # Assert that the sizes of the input and mask are the same.
    assert inputs.size() == mask.size(), f"Size of inputs {inputs.size()} and mask {mask.size()} expected to be the same"
    
    # Apply mask to inputs.
    return inputs * mask + value * (1 - mask)


"""# lucidium.spaces

Defines various space definitions for actions and observations.
"""

__all__ =   [
                # Abstract space.
                "Space",
                
                # Box space.
                "Box",
                
                # Composite space.
                "Composite",
                
                # Continuous spaces.
                "Continuous",
                "MultiContinuous",
                
                # Discrete spaces.
                "Discrete",
                "MultiDiscrete"
            ]

# Abstract space.
from lucidium.spaces.__base__           import Space

# Box space.
from lucidium.spaces.box                import Box

# Composite space.
from lucidium.spaces.composite          import Composite

# Continuous spaces.
from lucidium.spaces.continuous         import Continuous
from lucidium.spaces.multi_continuous   import MultiContinuous

# Discrete spaces.
from lucidium.spaces.discrete           import Discrete
from lucidium.spaces.multi_discrete     import MultiDiscrete


"""# lucidium.environments.tic_tac_toe.components

Modular components for Tic-Tac-Toe environment implementation.
"""

__all__ =   [
                "Board",
                "Cell",
                "Player"
            ]

from lucidium.environments.tic_tac_toe.components.board     import Board
from lucidium.environments.tic_tac_toe.components.cell      import Cell
from lucidium.environments.tic_tac_toe.components.players   import Player


"""# lucidium.agents.nlm.layers.inference

Defines the inference layers for neural logic machines.
"""

__all__ = ["InferenceLayer"]

from torch                          import Tensor
from torch.nn                       import Module, Sequential, Sigmoid

from lucidium.agents.nlm.layers.mlp import MLPLayer

class InferenceLayer(Module):
    """# Inference Layer

    MLP-based implementation for logic and logits inference.
    
    Adapted from: https://github.com/google/neural-logic-machines/blob/master/difflogic/nn/neural_logic/modules/neural_logic.py
    """
    
    def __init__(self,
        input_dimension:    int,
        output_dimension:   int,
        hidden_dimension:   int,
        activation:         bool =  False
    ):
        """# Initialize Inference Layer.

        ## Args:
            * input_dimensions  (list[int]):    The number of input channels for each input group.
            * output_dimensions (list[int]):    The number of output channels for each group.
            * hidden_dimension  (int):          The hidden dimension of the logic model.
            * activation        (bool):         Whether to add an activation layer after the MLP. Defaults to False.
        """
        # Initialize module.
        super(InferenceLayer, self).__init__()
        
        # Define dimensions.
        self._input_dimension_:     int =   input_dimension
        self._output_dimension_:    int =   output_dimension
        self._hidden_dimension_:    int =   hidden_dimension
        
        # Define layers.
        self._layer_:               Sequential =    Sequential(
                                                        MLPLayer(
                                                            input_dimension =   self._input_dimension_,
                                                            output_dimension =  self._output_dimension_,
                                                            hidden_dimensions = self._hidden_dimension_
                                                        )
                                                    )
        
        # Add activation layer if required.
        if activation: self._layer_.add_module(name = str(len(self._layer_)), module = Sigmoid())
        
    # PROPERTIES ===================================================================================
    
    @property
    def hidden_dimension(self) -> int:
        """# Hidden Dimension (int)
        
        The hidden dimension of the inference layer.
        """
        return self._hidden_dimension_
    
    @property
    def input_dimension(self) -> int:
        """# Input Dimension (int)
        
        The input dimension of the inference layer.
        """
        return self._input_dimension_
    
    @property
    def layer(self) -> Sequential:
        """# Layer (Sequential)
        
        The inference layer, which is a sequential module containing the MLP and optional 
        activation.
        """
        return self._layer_
    
    @property
    def output_dimension(self) -> int:
        """# Output Dimension (int)
        
        The output dimension of the inference layer.
        """
        return self._output_dimension_
    
    # METHODS ======================================================================================
        
    def forward(self,
        inputs: Tensor
    ) -> Tensor:
        """# Forward Pass.
        
        Make forward pass through inference layer.

        ## Args:
            * inputs    (Tensor):   Tensor input.

        ## Returns:
            * Tensor:   Output tensor.
        """
        # Reference input size and channels.
        size:       int =   inputs.size()[:-1]
        channel:    int =   inputs.size(-1)
        
        # Flatten input.
        inputs:     Tensor =    inputs.view(-1, channel)
        
        # Pass through layer.
        inputs:     Tensor =    self.layer(inputs)
        
        # Return flattened output.
        return inputs.view(*size, -1)


"""# lucidium.symbolic.predicate

Symbolic predicate definition, structure, and utilities.
"""

from typing import Any, Dict, Tuple

from torch  import tensor, Tensor

class Predicate():
    """# Predicate
    
    Symbolic representation of a grounded predicate.
    """
    
    def __init__(self,
        name:   str,
        *args:  Any
    ):
        """# Instantiate Predicate.

        ## Args:
            * name  (str):  Predicate name.
        """
        # Define properties.
        self._name_:    str =           name
        self._args_:    Tuple[Any] =    tuple(args)
        
    # PROPERTIES ===================================================================================
    
    @property
    def arity(self) -> int:
        """# (Predicate) Arity

        Number of arguments attrbuted to predicate.
        """
        return len(self._args_)
        
    # METHODS ======================================================================================
    
    def to_tensor(self,
        vocabulary: Dict[str, int]
    ) -> Tensor:
        """# (Predicate) to Tensor

        ## Args:
            * vocabulary    (Dict[str, int]):   Mapping of terms to their numeric value.

        ## Returns:
            * Tensor:   Tensor representation of predicate.
        """
        return tensor([vocabulary.get(str(arg), -1) for arg in (self._name_, *self._args_)])
    
    def to_tuple(self) -> Tuple[Any]:
        """# (Predicate) to Tuple

        Tuple representation of predicate.
        """
        return (self._name_, *self._args_)
    
    # DUNDERS ======================================================================================
    
    def __eq__(self,
        other: "Predicate"
    ) -> bool:
        """# Predicate Equality.

        ## Args:
            * other (Predicate):    Predicate being compared.

        ## Returns:
            * bool: True if predicates have the same name and arguments.
        """
        return isinstance(other, Predicate) and self.to_tuple() == other.to_tuple()
    
    def __repr__(self) -> str:
        """# Object Representation."""
        return f"""<Predicate(name = {self._name_}, arity = {self.arity})>"""
    
    def __str__(self) -> str:
        """# String Representation."""
        return f"""{self._name_}({",".join(map(str, self._args_))})"""